"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Optional

MONTHS = [
    "january", "february", "march", "april",
    "may", "june", "july", "august",
    "september", "october", "november", "december"
]


def map_month_name_to_month_number(month_name: str) -> Optional[int]:
    """
    Maps month names to month numbers
    :param month_name: The name of the month
    :return: The month number
    """
    month_name_mapping = {name: i + 1 for i, name in enumerate(MONTHS)}
    return month_name_mapping.get(month_name.lower())


def map_month_number_to_month_name(month_number: int) -> Optional[str]:
    """
    Maps month numbers to month names
    :param month_number: The month number
    :return: The name of the month
    """
    try:
        return MONTHS[month_number - 1]
    except IndexError:
        return None



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from puffotter.env import load_env_file
from puffotter.flask.initialize import init_flask
from puffotter.flask.wsgi import start_server
from otaku_info import sentry_dsn, root_path
from otaku_info.background import bg_tasks
from otaku_info.Config import Config
from otaku_info.routes import blueprint_generators
from otaku_info.db import models


def main():
    """
    Starts the flask application
    :return: None
    """
    load_env_file()
    init_flask(
        "otaku_info",
        sentry_dsn,
        root_path,
        Config,
        models,
        blueprint_generators
    )
    start_server(Config, bg_tasks)



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Dict, Any
from puffotter.flask.base import db
from puffotter.flask.db.ModelMixin import ModelMixin
from otaku_info.db.MediaUserState import MediaUserState


class MediaNotification(ModelMixin, db.Model):
    """
    Database model that stores a media notification for a user
    """

    __tablename__ = "media_notifications"
    """
    The name of the database table
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes the Model
        :param args: The constructor arguments
        :param kwargs: The constructor keyword arguments
        """
        super().__init__(*args, **kwargs)

    media_user_state_id: int = db.Column(
        db.Integer,
        db.ForeignKey(
            "media_user_states.id", ondelete="CASCADE", onupdate="CASCADE"
        ),
        nullable=False,
        unique=True
    )
    """
    The ID of the media user state this notification references
    """

    media_user_state: MediaUserState = db.relationship(
        "MediaUserState",
        backref=db.backref(
            "media_notifications", lazy=True, cascade="all,delete"
        )
    )
    """
    The media user state this notification references
    """

    last_update = db.Column(db.Integer, nullable=False)
    """
    The last update value sent to the user
    """

    def __json__(self, include_children: bool = False) -> Dict[str, Any]:
        """
        Generates a dictionary containing the information of this model
        :param include_children: Specifies if children data models
                                 will be included or if they're limited to IDs
        :return: A dictionary representing the model's values
        """
        data = {
            "id": self.id,
            "media_user_state_id": self.media_user_state_id,
            "last_update": self.last_update
        }
        if include_children:
            data["user"] = self.user.__json__(include_children)
            data["media_user_state"] = \
                self.media_user_state.__json__(include_children)
        return data



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Dict, Tuple, Callable
from otaku_info.background.anilist import fetch_anilist_data
from otaku_info.background.mangadex import load_id_mappings
from otaku_info.background.manga_chapters import \
    update_manga_chapter_guesses
from otaku_info.background.notifications import \
    send_new_manga_chapter_notifications
from otaku_info.background.ln_releases import update_ln_releases


bg_tasks: Dict[str, Tuple[int, Callable]] = {
    "anilist_update": (60, fetch_anilist_data),
    "update_manga_chapter_guesses": (60, update_manga_chapter_guesses),
    "load_id_mappings": (60 * 60 * 24, load_id_mappings),
    "manga_chapter_notifications": (60, send_new_manga_chapter_notifications),
    "ln_release_updates": (60 * 60 * 24, update_ln_releases)
}
"""
A dictionary containing background tasks for the flask application
"""



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from flask import url_for
from typing import Dict, Any
from puffotter.flask.base import db
from puffotter.flask.db.ModelMixin import ModelMixin
from otaku_info.utils.enums import ListService
from otaku_info.utils.db_model_helper import build_service_url
from otaku_info.db.MediaItem import MediaItem


class MediaId(ModelMixin, db.Model):
    """
    Database model for media IDs.
    These are used to map media items to their corresponding external
    IDS on external sites.
    """

    __tablename__ = "media_ids"
    """
    The name of the database table
    """

    __table_args__ = (
        db.UniqueConstraint(
            "media_item_id",
            "service",
            name="unique_media_id"
        ),
    )
    """
    Makes sure that objects that should be unique are unique
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes the Model
        :param args: The constructor arguments
        :param kwargs: The constructor keyword arguments
        """
        super().__init__(*args, **kwargs)

    media_item_id: int = db.Column(
        db.Integer,
        db.ForeignKey(
            "media_items.id", ondelete="CASCADE", onupdate="CASCADE"
        ),
        nullable=False
    )
    """
    The ID of the media item referenced by this ID
    """

    media_item: MediaItem = db.relationship(
        "MediaItem",
        backref=db.backref("media_ids", lazy=True, cascade="all,delete")
    )
    """
    The media item referenced by this ID
    """

    service_id: str = db.Column(db.String(255), nullable=False)
    """
    The ID of the media item on the external service
    """

    service: ListService = db.Column(db.Enum(ListService), nullable=False)
    """
    The service for which this object represents an ID
    """

    @property
    def service_url(self) -> str:
        """
        :return: The URL to the series for the given service
        """
        return build_service_url(
            self.media_item.media_type, self.service, self.service_id
        )

    @property
    def service_icon(self) -> str:
        """
        :return: The path to the service's icon file
        """
        return url_for(
            "static", filename="service_logos/" + self.service.value + ".png"
        )

    def __json__(self, include_children: bool = False) -> Dict[str, Any]:
        """
        Generates a dictionary containing the information of this model
        :param include_children: Specifies if children data models
                                 will be included or if they're limited to IDs
        :return: A dictionary representing the model's values
        """
        data = {
            "id": self.id,
            "media_item_id": self.media_item_id,
            "service_id": self.service_id,
            "service": self.service.value
        }
        if include_children:
            data["media_item"] = self.media_item.__json__(include_children)
        return data



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Tuple
from puffotter.flask.base import db
from puffotter.flask.db.User import User
from sqlalchemy.exc import IntegrityError
from otaku_info.db.MediaItem import MediaItem
from otaku_info.db.MediaId import MediaId
from otaku_info.db.MediaUserState import MediaUserState
from otaku_info.utils.enums import ListService, MediaType, MediaSubType, \
    ReleasingState, ConsumingState
from otaku_info.test.TestFramework import _TestFramework


class TestMediaUserState(_TestFramework):
    """
    Class that tests the MediaUserState database model
    """

    def generate_sample_media_user_state(self) \
            -> Tuple[MediaUserState, User, MediaItem, MediaId]:
        """
        Generates a media user state
        :return: The media user state, user, media item and media id
        """
        user = self.generate_sample_user(True)[0]
        media_item = MediaItem(
            media_type=MediaType.MANGA,
            media_subtype=MediaSubType.MANGA,
            english_title="Fly Me to the Moon",
            romaji_title="Tonikaku Cawaii",
            cover_url="https://s4.anilist.co/file/anilistcdn/media/manga/"
                      "cover/medium/nx101177-FjjD5UWB3C3t.png",
            latest_release=None,
            releasing_state=ReleasingState.RELEASING
        )
        media_id = MediaId(
            media_item=media_item,
            service_id="101177",
            service=ListService.ANILIST
        )
        media_user_state = MediaUserState(
            media_id=media_id,
            user=user,
            progress=10,
            score=75,
            consuming_state=ConsumingState.CURRENT
        )
        db.session.add(media_item)
        db.session.add(media_id)
        db.session.add(media_user_state)
        db.session.commit()
        return media_user_state, user, media_item, media_id

    def test_json_representation(self):
        """
        Tests the JSON representation of the model
        :return: None
        """
        media_user_state, user, media_item, media_id = \
            self.generate_sample_media_user_state()

        self.assertEqual(
            media_user_state.__json__(False),
            {
                "id": media_id.id,
                "media_id_id": media_id.id,
                "user_id": user.id,
                "progress": media_user_state.progress,
                "score": media_user_state.score,
                "consuming_state": media_user_state.consuming_state.value
            }
        )
        self.assertEqual(
            media_user_state.__json__(True),
            {
                "id": media_id.id,
                "media_id": media_id.__json__(True),
                "media_id_id": media_id.id,
                "user_id": user.id,
                "user": user.__json__(True),
                "progress": media_user_state.progress,
                "score": media_user_state.score,
                "consuming_state": media_user_state.consuming_state.value
            }
        )

    def test_string_representation(self):
        """
        Tests the string representation of the model
        :return: None
        """
        media_user_state, user, media_item, media_id = \
            self.generate_sample_media_user_state()
        data = media_user_state.__json__()
        data.pop("id")
        self.assertEqual(
            str(media_user_state),
            "MediaUserState:{} <{}>".format(media_user_state.id, data)
        )

    def test_repr(self):
        """
        Tests the __repr__ method of the model class
        :return: None
        """
        media_user_state, user, media_item, media_id = \
            self.generate_sample_media_user_state()
        generated = {"value": media_user_state}
        code = repr(media_user_state)

        exec("generated['value'] = " + code)
        self.assertEqual(generated["value"], media_user_state)
        self.assertFalse(generated["value"] is media_user_state)

    def test_hashing(self):
        """
        Tests using the model objects as keys in a dictionary
        :return: None
        """
        media_user_state, user, media_item, media_id = \
            self.generate_sample_media_user_state()
        media_user_state_2 = MediaUserState(
            media_id=media_id,
            user=self.generate_sample_user()[0],
            progress=15,
            score=85,
            consuming_state=ConsumingState.CURRENT
        )
        db.session.add(media_user_state_2)
        db.session.commit()
        mapping = {
            media_user_state: 100,
            media_user_state_2: 200
        }
        self.assertEqual(mapping[media_user_state], 100)
        self.assertEqual(mapping[media_user_state_2], 200)

    def test_equality(self):
        """
        Tests checking equality for model objects
        :return: None
        """
        media_user_state, user, media_item, media_id = \
            self.generate_sample_media_user_state()
        media_user_state_2 = MediaUserState(
            media_id=media_id,
            user=self.generate_sample_user()[0],
            progress=15,
            score=85,
            consuming_state=ConsumingState.CURRENT
        )
        db.session.add(media_user_state_2)
        db.session.commit()
        self.assertEqual(media_user_state, media_user_state)
        self.assertNotEqual(media_user_state, media_user_state_2)
        self.assertNotEqual(media_user_state, 100)

    def test_uniqueness(self):
        """
        Tests if the uniqueness of the model is handled properly
        :return: None
        """
        media_user_state, user, media_item, media_id = \
            self.generate_sample_media_user_state()

        standard_kwargs = media_user_state.__json__(False)
        standard_kwargs.pop("id")
        standard_kwargs["consuming_state"] = media_user_state.consuming_state

        media_id_kwargs = media_id.__json__(False)
        media_id_kwargs.pop("id")
        media_id_kwargs["service"] = ListService.KITSU
        new_media_id = MediaId(**media_id_kwargs)
        db.session.add(new_media_id)
        db.session.commit()

        try:
            duplicate = MediaUserState(**standard_kwargs)
            db.session.add(duplicate)
            db.session.commit()
            self.fail()
        except IntegrityError:
            db.session.rollback()

        for key, value, error_expected in [
            ("media_id_id", new_media_id.id, False),
            ("user_id", self.generate_sample_user(True)[0].id, False),
            ("progress", 1, True),
            ("score", 12, True),
            ("consuming_state", ConsumingState.PAUSED, True)
        ]:
            kwargs = dict(standard_kwargs)
            kwargs[key] = value
            try:
                generated = MediaUserState(**kwargs)
                db.session.add(generated)
                db.session.commit()
                if error_expected:
                    self.fail()
            except IntegrityError as e:
                db.session.rollback()
                if not error_expected:
                    raise e



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import time
from puffotter.flask.base import db, app
from otaku_info.db.MediaUserState import MediaUserState
from otaku_info.db.MangaChapterGuess import MangaChapterGuess
from otaku_info.utils.enums import MediaType, ListService


def update_manga_chapter_guesses():
    """
    Updates the manga chapter guesses
    :return: None
    """
    app.logger.debug("Starting update of manga chapter guesses")
    anilist_ids = {
        x.media_id.service_id: x.media_id
        for x in MediaUserState.query.all()
        if x.media_id.media_item.media_type == MediaType.MANGA
        and x.media_id.service == ListService.ANILIST
    }
    guesses = {
        x.media_id.service_id: x
        for x in MangaChapterGuess.query.all()
    }

    for anilist_id in anilist_ids:
        if anilist_id not in guesses:
            new_guess = MangaChapterGuess(media_id=anilist_ids[anilist_id])
            db.session.add(new_guess)
            guesses[anilist_id] = new_guess

    db.session.commit()

    for anilist_id, guess in guesses.items():

        if anilist_id not in anilist_ids:
            db.session.delete(guess)
            app.logger.debug(f"Deleting stale chapter guess for {anilist_id}")
        else:
            app.logger.debug(f"Updating chapter guess for {anilist_id}")
            guess.update()
            time.sleep(1)

        db.session.commit()

    app.logger.debug("Finished updating manga chapter guesses")



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Dict, List
from puffotter.flask.base import db
from otaku_info.Config import Config
from puffotter.flask.db.TelegramChatId import TelegramChatId
from otaku_info.db.MediaUserState import MediaUserState
from otaku_info.db.MediaNotification import MediaNotification
from otaku_info.db.MangaChapterGuess import MangaChapterGuess
from otaku_info.db.MediaId import MediaId
from otaku_info.db.MediaItem import MediaItem
from otaku_info.db.NotificationSetting import NotificationSetting
from otaku_info.utils.enums import \
    MediaType, MediaSubType, ConsumingState, NotificationType


def send_new_manga_chapter_notifications():
    """
    Sends out telegram notifications for manga chapter updates
    :return: None
    """

    chats: Dict[int, TelegramChatId] = {
        x.user_id: x for x in TelegramChatId.query.all()
    }
    chapter_guesses: Dict[int, int] = {
        x.media_id_id: x.guess for x in MangaChapterGuess.query.all()
    }
    user_states: List[MediaUserState] = MediaUserState.query\
        .join(MediaId) \
        .join(MediaItem) \
        .filter(MediaItem.media_type == MediaType.MANGA)\
        .filter(MediaItem.media_subtype == MediaSubType.MANGA)\
        .filter(MediaUserState.consuming_state == ConsumingState.CURRENT)\
        .all()
    notifications: Dict[int, MediaNotification] = {
        x.media_user_state_id: x for x in MediaNotification.query.all()
    }
    notification_settings: Dict[int, NotificationSetting] = {
        x.user_id: x
        for x in NotificationSetting.query.filter_by(
            notification_type=NotificationType.NEW_MANGA_CHAPTERS
        ).all()
    }

    for user_state in user_states:

        settings = notification_settings.get(user_state.user_id)
        if settings is None or not settings.value:
            continue

        guess = chapter_guesses.get(user_state.media_id_id)
        notification = notifications.get(user_state.id)
        chat = chats.get(user_state.user_id)
        progress = user_state.progress

        if guess is None or chat is None:
            continue

        diff = guess - progress

        if diff <= 0:
            continue

        if notification is None:
            notification = MediaNotification(
                media_user_state=user_state, last_update=guess
            )
            db.session.add(notification)

        if guess != notification.last_update:

            notification.last_update = guess
            if user_state.score >= settings.minimum_score:

                title = user_state.media_id.media_item.title
                url = f"https://{Config.DOMAIN_NAME}/media/" \
                      f"{user_state.media_id.media_item_id}"
                chat.send_message(
                    f"New Chapter for {title}\n\n"
                    f"Chapter {progress}/{guess} (+{diff})\n\n"
                    f"{url}"
                )

    db.session.commit()



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import List
from puffotter.flask.base import db
from otaku_info.db.MangaChapterGuess import MangaChapterGuess
from otaku_info.db.MediaId import MediaId
from otaku_info.db.MediaItem import MediaItem
from otaku_info.db.MediaList import MediaList
from otaku_info.db.MediaListItem import MediaListItem
from otaku_info.db.MediaUserState import MediaUserState
from otaku_info.db.MediaNotification import MediaNotification
from otaku_info.db.ServiceUsername import ServiceUsername
from otaku_info.db.NotificationSetting import NotificationSetting
from otaku_info.db.LnRelease import LnRelease

models: List[db.Model] = [
    MangaChapterGuess,
    MediaId,
    MediaItem,
    MediaList,
    MediaListItem,
    MediaUserState,
    ServiceUsername,
    MediaNotification,
    NotificationSetting,
    LnRelease
]
"""
The database models of the application
"""



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import time
from typing import Dict, List, Optional, Tuple
from sqlalchemy.exc import IntegrityError
from puffotter.flask.base import db, app
from otaku_info.db.MediaItem import MediaItem
from otaku_info.db.MediaId import MediaId
from otaku_info.utils.enums import ListService, MediaType
from otaku_info.utils.mangadex.api import get_external_ids
from otaku_info.utils.anilist.api import load_media_info


def load_id_mappings():
    """
    Goes through mangadex IDs sequentially and stores ID mappings for
    these entries if found
    :return: None
    """
    endcounter = 0
    mangadex_id = 0

    anilist_ids, existing_ids = load_db_content()

    while True:
        mangadex_id += 1

        if mangadex_id % 100 == 0:
            app.logger.debug("Refreshing mangadex cache")
            anilist_ids, existing_ids = load_db_content()

        app.logger.debug(f"Probing mangadex id {mangadex_id}")

        other_ids = get_external_ids(mangadex_id)

        if other_ids is None:
            endcounter += 1
            if endcounter > 1000:
                break
            else:
                continue
        else:
            endcounter = 0

        store_ids(existing_ids, anilist_ids, mangadex_id, other_ids)
    app.logger.info("Reached end of mangadex ID range")


def load_db_content() -> Tuple[
    Dict[str, MediaId],
    Dict[int, List[ListService]]
]:
    """
    Loads the existing data from the database.
    By doing this as few times as possible, we can greatly improve performance
    :return: The anilist IDs, The mangadex IDs mapped to other existing IDs
    """
    start = time.time()
    app.logger.debug("Starting caching of db data for mangadex ID mapping")

    all_ids: List[MediaId] = [
        x for x in
        MediaId.query.join(MediaItem).all()
        if x.media_item.media_type == MediaType.MANGA
    ]
    anilist_ids: Dict[str, MediaId] = {
        x.service_id: x
        for x in all_ids
        if x.service == ListService.ANILIST
    }

    mangadex_idmap: Dict[int, int] = {}

    existing_ids: Dict[int, List[ListService]] = {}
    for media_id in all_ids:
        media_item_id = media_id.media_item_id
        if media_item_id not in existing_ids:
            existing_ids[media_item_id] = []
        existing_ids[media_item_id].append(media_id)
        if media_id.service == ListService.MANGADEX:
            mangadex_idmap[media_item_id] = int(media_id.service_id)

    mapped_existing_ids = {
        mangadex_idmap[key]: value
        for key, value in existing_ids.items()
        if key in mangadex_idmap
    }

    app.logger.info(f"Finished caching of db data for mangadex ID mapping "
                    f"in {time.time() - start}s")
    return anilist_ids, mapped_existing_ids


def store_ids(
        existing_ids: Dict[int, List[ListService]],
        anilist_ids: Dict[str, MediaId],
        mangadex_id: int,
        other_ids: Dict[ListService, str]
):
    """
    Stores the fetched IDs in the database
    :param existing_ids: A dictionary mapping mangadex IDs to existing
                         list service types
    :param anilist_ids: Dictionary mapping anilist IDs to media IDs
    :param mangadex_id: The mangadex ID
    :param other_ids: The other IDs
    :return: None
    """
    if ListService.ANILIST not in other_ids:
        return

    existing_services = existing_ids.get(mangadex_id, [])
    existing_ids[mangadex_id] = existing_services
    anilist_id = other_ids[ListService.ANILIST]

    if anilist_id not in anilist_ids:
        media_item = create_anilist_media_item(int(anilist_id))
        if media_item is None:
            return
        else:
            media_item_id = media_item.id
    else:
        media_item_id = anilist_ids[anilist_id].media_item_id
        existing_services.append(ListService.ANILIST)

    app.logger.debug(f"Storing external IDS for mangadex id {mangadex_id}")

    for list_service, _id in other_ids.items():
        if list_service not in existing_services:
            media_id = MediaId(
                media_item_id=media_item_id,
                service=list_service,
                service_id=_id
            )
            db.session.add(media_id)
            existing_ids[mangadex_id].append(list_service)
            if list_service == ListService.ANILIST:
                anilist_ids[_id] = media_id

    try:
        db.session.commit()
    except IntegrityError:
        # Since mangadex has some entries that point to the exact same anilist
        # media item, we may sometimes encounter cases where we have two
        # mangadex IDs for one anilist ID.
        # By ignoring errors here, only the first mangadex ID will be stored.
        # An example for this issue is Genshiken (961) and its
        # sequel Genshiken Nidaime (962)
        db.session.rollback()
        app.logger.warning(f"Couldn't add mangadex ID {mangadex_id}")


def create_anilist_media_item(anilist_id: int) -> Optional[MediaItem]:
    """
    Creates an anilist media item using an anilist ID, fetching the data using
    the anilist API
    :param anilist_id: The anilist ID of the media
    :return: The generated Media Item
    """
    anilist_entry = load_media_info(anilist_id, MediaType.MANGA)
    if anilist_entry is None:
        return None
    media_item = MediaItem(
        media_type=MediaType.MANGA,
        media_subtype=anilist_entry.media_subtype,
        english_title=anilist_entry.english_title,
        romaji_title=anilist_entry.romaji_title,
        cover_url=anilist_entry.cover_url,
        latest_release=anilist_entry.latest_release,
        releasing_state=anilist_entry.releasing_state
    )
    db.session.add(media_item)

    try:
        db.session.commit()
    except IntegrityError:
        db.session.rollback()
        app.logger.warning(f"Failed to add anilist manga entry "
                           f"(ID={anilist_id})")
        return None

    return media_item



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Dict, Any
from flask_login import current_user
from otaku_info.utils.enums import ListService
from otaku_info.db.ServiceUsername import ServiceUsername


def profile_extras() -> Dict[str, Any]:
    """
    Makes sure that the profile page displays service usernames
    :return: The variables to forward to the template
    """
    service_usernames = {}
    for service in ListService:
        service_username = ServiceUsername.query.filter_by(
            user=current_user, service=service
        ).first()
        service_usernames[service] = service_username
    return {
        "service_usernames": service_usernames
    }



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Tuple
from puffotter.flask.base import db
from puffotter.flask.db.User import User
from sqlalchemy.exc import IntegrityError
from otaku_info.db.MediaList import MediaList
from otaku_info.utils.enums import MediaType, ListService
from otaku_info.test.TestFramework import _TestFramework


class TestMediaList(_TestFramework):
    """
    Class that tests the MediaList database model
    """

    def generate_sample_media_list(self) -> Tuple[MediaList, User]:
        """
        Generates a sample media item
        :return: The media list and the associated user
        """
        user = self.generate_sample_user(True)[0]
        media_list = MediaList(
            name="ABC",
            user=user,
            service=ListService.ANILIST,
            media_type=MediaType.MANGA
        )
        db.session.add(media_list)
        db.session.commit()
        return media_list, user

    def test_json_representation(self):
        """
        Tests the JSON representation of the model
        :return: None
        """
        media_list, user = self.generate_sample_media_list()

        self.assertEqual(
            media_list.__json__(False),
            {
                "id": media_list.id,
                "name": media_list.name,
                "user_id": user.id,
                "service": media_list.service.value,
                "media_type": media_list.media_type.value
            }
        )
        self.assertEqual(
            media_list.__json__(True),
            {
                "id": media_list.id,
                "name": media_list.name,
                "user_id": user.id,
                "user": user.__json__(True),
                "service": media_list.service.value,
                "media_type": media_list.media_type.value
            }
        )

    def test_string_representation(self):
        """
        Tests the string representation of the model
        :return: None
        """
        media_list, user = self.generate_sample_media_list()
        data = media_list.__json__()
        data.pop("id")
        self.assertEqual(
            str(media_list),
            "MediaList:{} <{}>".format(media_list.id, data)
        )

    def test_repr(self):
        """
        Tests the __repr__ method of the model class
        :return: None
        """
        media_list, user = self.generate_sample_media_list()
        generated = {"value": media_list}
        code = repr(media_list)

        exec("generated['value'] = " + code)
        self.assertEqual(generated["value"], media_list)
        self.assertFalse(generated["value"] is media_list)

    def test_hashing(self):
        """
        Tests using the model objects as keys in a dictionary
        :return: None
        """
        media_list, user = self.generate_sample_media_list()
        media_list_2 = MediaList(
            name="XYZ",
            user=user,
            service=ListService.ANILIST,
            media_type=MediaType.MANGA
        )
        db.session.add(media_list_2)
        db.session.commit()
        mapping = {
            media_list: 100,
            media_list_2: 200
        }
        self.assertEqual(mapping[media_list], 100)
        self.assertEqual(mapping[media_list_2], 200)

    def test_equality(self):
        """
        Tests checking equality for model objects
        :return: None
        """
        media_list, user = self.generate_sample_media_list()
        media_list_2 = MediaList(
            name="XYZ",
            user=user,
            service=ListService.ANILIST,
            media_type=MediaType.MANGA
        )
        db.session.add(media_list_2)
        db.session.commit()
        self.assertEqual(media_list, media_list)
        self.assertNotEqual(media_list, media_list_2)
        self.assertNotEqual(media_list, 100)

    def test_uniqueness(self):
        """
        Tests if the uniqueness of the model is handled properly
        :return: None
        """
        media_list, user = self.generate_sample_media_list()
        standard_kwargs = media_list.__json__(False)
        standard_kwargs.pop("id")
        standard_kwargs["service"] = media_list.service
        standard_kwargs["media_type"] = media_list.media_type

        try:
            duplicate = MediaList(**standard_kwargs)
            db.session.add(duplicate)
            db.session.commit()
            self.fail()
        except IntegrityError:
            db.session.rollback()

        for key, value in [
            ("name", "ABABABABA"),
            ("user_id", self.generate_sample_user(True)[0].id),
            ("service", ListService.ANIMEPLANET),
            ("media_type", MediaType.ANIME)
        ]:
            kwargs = dict(standard_kwargs)
            kwargs[key] = value

            generated = MediaList(**kwargs)
            db.session.add(generated)
            db.session.commit()



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from flask import url_for
from typing import Optional, List, Tuple, Dict, Any
from otaku_info.utils.enums import ListService, MediaType
from otaku_info.utils.db_model_helper import build_service_url


class MangaUpdate:
    """
    Class that encapsulates important data to display for manga updates
    """

    def __init__(
            self,
            media_item_id: int,
            title: str,
            cover_url: str,
            latest_release: int,
            progress: int,
            score: int,
            chapter_guess: Optional[int],
            related_ids: List[Tuple[ListService, str]]
    ):
        """
        Initializes the MangaUpdate object
        :param media_item_id: The media item ID
        :param title: The title of the update
        :param cover_url: The URL for the media item's cover
        :param latest_release: The latest known released chapter
        :param progress: The user's current progress
        :param score: The user's score for this entry
        :param chapter_guess: The current chapter guess
        :param related_ids: Related service IDs
        """
        self.media_item_id = media_item_id
        self.title = title
        self.cover_url = cover_url
        self.score = score
        self.progress = progress
        self.related_ids = [RelatedMangaId(*args) for args in related_ids]
        self.url = url_for("media.media", media_item_id=media_item_id)

        if chapter_guess is None:
            self.latest = latest_release
        else:
            self.latest = chapter_guess

        if self.latest is None:
            self.latest = 0
        if self.progress is None:
            self.progress = 0

        if self.latest < self.progress:
            self.latest = self.progress

        self.diff = self.latest - self.progress

    def __json__(self) -> Dict[str, Any]:
        """
        Converts the object into a JSON-compatible dictionary
        :return: The JSON-compatible dictionary
        """
        return {
            "title": self.title,
            "cover_url": self.cover_url,
            "score": self.score,
            "progress": self.progress,
            "related_ids": [
                {
                    "service": x.service.value,
                    "id": x.service_id,
                    "url": x.url,
                    "icon": x.icon
                }
                for x in self.related_ids
            ],
            "latest": self.latest,
            "diff": self.diff
        }


class RelatedMangaId:
    """
    Class that encapslates attributes for a related manga ID
    """

    def __init__(self, service: ListService, service_id: str):
        """
        Intializes the RelatedMangaId object
        :param service: The service of the related manga ID
        :param service_id: The ID on that service
        """
        self.service = service
        self.service_id = service_id
        self.url = build_service_url(MediaType.MANGA, service, service_id)
        self.icon = url_for(
            "static", filename="service_logos/" + service.value + ".png"
        )



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from flask import request, render_template, redirect, url_for
from flask.blueprints import Blueprint
from flask_login import login_required, current_user
from otaku_info.utils.enums import MediaType, ListService
from otaku_info.utils.manga_updates.generator import prepare_manga_updates
from otaku_info.db.MediaList import MediaList


def define_blueprint(blueprint_name: str) -> Blueprint:
    """
    Defines the blueprint for this route
    :param blueprint_name: The name of the blueprint
    :return: The blueprint
    """
    blueprint = Blueprint(blueprint_name, __name__)

    @blueprint.route("/manga/updates", methods=["POST"])
    @login_required
    def redirect_manga_updates():
        """
        Redirects a POST requests to the appropriate GET request for
        the /manga/updates route
        :return: The response
        """
        service, list_name = request.form["list_ident"].split(":", 1)
        mincount = request.form.get("mincount", "0")
        include_complete = request.form.get("include_complete", "off") == "on"

        get_url = url_for("manga.show_manga_updates")
        get_url += f"?service={service}" \
                   f"&list_name={list_name}" \
                   f"&mincount={mincount}" \
                   f"&include_complete={1 if include_complete else 0}"

        return redirect(get_url)

    @blueprint.route("/manga/updates", methods=["GET"])
    @login_required
    def show_manga_updates():
        """
        Shows the user's manga updates for a specified service and list
        :return: The response
        """
        service = request.args.get("service")
        list_name = request.args.get("list_name")
        mincount = int(request.args.get("mincount", "0"))
        include_complete = request.args.get("include_complete", "0") == "1"

        if service is None or list_name is None:
            media_lists = [
                x for x in MediaList.query.filter_by(
                    user=current_user, media_type=MediaType.MANGA
                )
            ]
            return render_template(
                "manga/manga_updates.html",
                media_lists=media_lists
            )
        else:
            list_entries = \
                prepare_manga_updates(
                    current_user,
                    ListService(service),
                    list_name,
                    include_complete,
                    mincount
                )
            list_entries.sort(key=lambda x: x.score, reverse=True)
            return render_template(
                "manga/manga_updates.html",
                entries=list_entries,
                list_name=list_name,
                service=service
            )

    return blueprint



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Tuple
from puffotter.flask.base import db
from puffotter.flask.db.User import User
from sqlalchemy.exc import IntegrityError
from otaku_info.db.ServiceUsername import ServiceUsername
from otaku_info.utils.enums import ListService
from otaku_info.test.TestFramework import _TestFramework


class TestServiceUsername(_TestFramework):
    """
    Class that tests the ServiceUsername database model
    """

    def generate_sample_service_username(self) -> Tuple[ServiceUsername, User]:
        """
        Generates a sample service username
        :return: The service username and the user
        """
        user = self.generate_sample_user(True)[0]
        service_username = ServiceUsername(
            user=user,
            username="abc",
            service=ListService.ANILIST
        )
        db.session.add(service_username)
        db.session.commit()
        return service_username, user

    def test_json_representation(self):
        """
        Tests the JSON representation of the model
        :return: None
        """
        service_username, user = self.generate_sample_service_username()

        self.assertEqual(
            service_username.__json__(False),
            {
                "id": service_username.id,
                "user_id": user.id,
                "username": service_username.username,
                "service": service_username.service.value
            }
        )
        self.assertEqual(
            service_username.__json__(True),
            {
                "id": service_username.id,
                "user_id": user.id,
                "user": user.__json__(True),
                "username": service_username.username,
                "service": service_username.service.value
            }
        )

    def test_string_representation(self):
        """
        Tests the string representation of the model
        :return: None
        """
        service_username, user = self.generate_sample_service_username()
        data = service_username.__json__()
        data.pop("id")
        self.assertEqual(
            str(service_username),
            "ServiceUsername:{} <{}>".format(service_username.id, data)
        )

    def test_repr(self):
        """
        Tests the __repr__ method of the model class
        :return: None
        """
        service_username, user = self.generate_sample_service_username()
        generated = {"value": service_username}
        code = repr(service_username)

        exec("generated['value'] = " + code)
        self.assertEqual(generated["value"], service_username)
        self.assertFalse(generated["value"] is service_username)

    def test_hashing(self):
        """
        Tests using the model objects as keys in a dictionary
        :return: None
        """
        service_username, user = self.generate_sample_service_username()
        service_username_2 = ServiceUsername(
            user=user,
            username="xyz",
            service=ListService.MYANIMELIST
        )
        db.session.add(service_username_2)
        db.session.commit()
        mapping = {
            service_username: 100,
            service_username_2: 200
        }
        self.assertEqual(mapping[service_username], 100)
        self.assertEqual(mapping[service_username_2], 200)

    def test_equality(self):
        """
        Tests checking equality for model objects
        :return: None
        """
        service_username, user = self.generate_sample_service_username()
        service_username_2 = ServiceUsername(
            user=user,
            username="xyz",
            service=ListService.MYANIMELIST
        )
        db.session.add(service_username_2)
        db.session.commit()
        self.assertEqual(service_username, service_username)
        self.assertNotEqual(service_username, service_username_2)
        self.assertNotEqual(service_username, 100)

    def test_uniqueness(self):
        """
        Tests if the uniqueness of the model is handled properly
        :return: None
        """
        service_username, user = self.generate_sample_service_username()

        standard_kwargs = service_username.__json__(False)
        standard_kwargs.pop("id")
        standard_kwargs["service"] = service_username.service

        try:
            duplicate = ServiceUsername(**standard_kwargs)
            db.session.add(duplicate)
            db.session.commit()
            self.fail()
        except IntegrityError:
            db.session.rollback()

        for key, value in [
            ("user_id", self.generate_sample_user()[0].id),
            ("username", "NEW"),
            ("service", ListService.KITSU)
        ]:
            kwargs = dict(standard_kwargs)
            kwargs[key] = value

            generated = ServiceUsername(**kwargs)
            db.session.add(generated)
            db.session.commit()



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Tuple
from unittest.mock import patch
from puffotter.flask.base import db
from sqlalchemy.exc import IntegrityError
from otaku_info.db.MediaItem import MediaItem
from otaku_info.db.MediaId import MediaId
from otaku_info.db.MangaChapterGuess import MangaChapterGuess
from otaku_info.utils.enums import ListService, MediaType, MediaSubType, \
    ReleasingState
from otaku_info.test.TestFramework import _TestFramework


class TestMangaChapterGuess(_TestFramework):
    """
    Class that tests the MangaChapterGuess database model
    """

    @staticmethod
    def generate_sample_guess() \
            -> Tuple[MediaItem, MediaId, MangaChapterGuess]:
        """
        Generates a sample chapter guess
        :return: The media item, media id and chapter guess
        """
        media_item = MediaItem(
            media_type=MediaType.MANGA,
            media_subtype=MediaSubType.MANGA,
            english_title="Fly Me to the Moon",
            romaji_title="Tonikaku Cawaii",
            cover_url="https://s4.anilist.co/file/anilistcdn/media/manga/"
                      "cover/medium/nx101177-FjjD5UWB3C3t.png",
            latest_release=None,
            releasing_state=ReleasingState.RELEASING
        )
        media_id = MediaId(
            media_item=media_item,
            service_id="101177",
            service=ListService.ANILIST
        )
        chapter_guess = MangaChapterGuess(
            media_id=media_id,
            guess=None,
            last_update=0
        )
        db.session.add(media_item)
        db.session.add(media_id)
        db.session.add(chapter_guess)
        db.session.commit()
        return media_item, media_id, chapter_guess

    def test_json_representation(self):
        """
        Tests the JSON representation of the model
        :return: None
        """
        _, media_id, chapter_guess = self.generate_sample_guess()

        self.assertEqual(
            chapter_guess.__json__(False),
            {
                "id": chapter_guess.id,
                "media_id_id": media_id.id,
                "guess": chapter_guess.guess,
                "last_update": chapter_guess.last_update
            }
        )
        self.assertEqual(
            chapter_guess.__json__(True),
            {
                "id": chapter_guess.id,
                "media_id": media_id.__json__(True),
                "media_id_id": media_id.id,
                "guess": chapter_guess.guess,
                "last_update": chapter_guess.last_update
            }
        )

    def test_string_representation(self):
        """
        Tests the string representation of the model
        :return: None
        """
        media_item, media_id, chapter_guess = self.generate_sample_guess()
        data = chapter_guess.__json__()
        data.pop("id")
        self.assertEqual(
            str(chapter_guess),
            "MangaChapterGuess:{} <{}>".format(chapter_guess.id, data)
        )

    def test_repr(self):
        """
        Tests the __repr__ method of the model class
        :return: None
        """
        _, __, chapter_guess = self.generate_sample_guess()
        generated = {"value": chapter_guess}
        code = repr(chapter_guess)

        exec("generated['value'] = " + code)
        self.assertEqual(generated["value"], chapter_guess)
        self.assertFalse(generated["value"] is chapter_guess)

    def test_hashing(self):
        """
        Tests using the model objects as keys in a dictionary
        :return: None
        """
        media_item, media_id, chapter_guess = self.generate_sample_guess()

        media_id_kwargs = media_id.__json__(False)
        media_id_kwargs.pop("id")
        media_id_kwargs["service"] = ListService.ANIMEPLANET
        new_media_id = MediaId(**media_id_kwargs)
        db.session.add(new_media_id)
        db.session.commit()

        chapter_guess_2 = MangaChapterGuess(
            media_id=new_media_id,
            guess=100,
            last_update=0
        )
        db.session.add(chapter_guess_2)
        db.session.commit()
        mapping = {
            chapter_guess: 100,
            chapter_guess_2: 200
        }
        self.assertEqual(mapping[chapter_guess], 100)
        self.assertEqual(mapping[chapter_guess_2], 200)

    def test_equality(self):
        """
        Tests checking equality for model objects
        :return: None
        """
        _, media_id, chapter_guess = self.generate_sample_guess()

        media_id_kwargs = media_id.__json__(False)
        media_id_kwargs.pop("id")
        media_id_kwargs["service"] = ListService.ANIMEPLANET
        new_media_id = MediaId(**media_id_kwargs)
        db.session.add(new_media_id)
        db.session.commit()

        chapter_guess_2 = MangaChapterGuess(
            media_id=new_media_id,
            guess=100,
            last_update=0
        )
        self.assertEqual(chapter_guess, chapter_guess)
        self.assertNotEqual(chapter_guess, chapter_guess_2)
        self.assertNotEqual(chapter_guess, 100)

    def test_guessing_latest_chapter(self):
        """
        Tests guessing the latest chapter
        :return: None
        """
        _, _, chapter_guess = self.generate_sample_guess()

        with patch("otaku_info.db.MangaChapterGuess"
                   ".guess_latest_manga_chapter", lambda _: 105):
            self.assertEqual(chapter_guess.guess, None)
            self.assertEqual(chapter_guess.last_update, 0)
            chapter_guess.update()
            self.assertEqual(chapter_guess.guess, 105)
            self.assertGreater(chapter_guess.last_update, 0)
            chapter_guess.guess = 77
            last_update = chapter_guess.last_update
            chapter_guess.update()
            self.assertEqual(chapter_guess.guess, 77)
            self.assertEqual(chapter_guess.last_update, last_update)
            chapter_guess.last_update = 0
            chapter_guess.update()
            self.assertEqual(chapter_guess.guess, 105)
            self.assertGreaterEqual(chapter_guess.last_update, last_update)

    def test_uniqueness(self):
        """
        Tests if the uniqueness of the model is handled properly
        :return: None
        """
        _, media_id, chapter_guess = self.generate_sample_guess()
        duplicate = MangaChapterGuess(
            media_id_id=media_id.id,
            guess=12345,
            last_update=12345
        )
        try:
            db.session.add(duplicate)
            db.session.commit()
            self.fail()
        except IntegrityError:
            pass



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from sqlalchemy.exc import IntegrityError
from puffotter.flask.base import db
from otaku_info.db.MediaItem import MediaItem
from otaku_info.utils.enums import MediaType, MediaSubType, ReleasingState
from otaku_info.test.TestFramework import _TestFramework


class TestMediaItem(_TestFramework):
    """
    Class that tests the MediaItem database model
    """

    @staticmethod
    def generate_sample_media_item() -> MediaItem:
        """
        Generates a sample media item
        :return: The media item
        """
        media_item = MediaItem(
            media_type=MediaType.MANGA,
            media_subtype=MediaSubType.MANGA,
            english_title="Fly Me to the Moon",
            romaji_title="Tonikaku Cawaii",
            cover_url="https://s4.anilist.co/file/anilistcdn/media/manga/"
                      "cover/medium/nx101177-FjjD5UWB3C3t.png",
            latest_release=None,
            releasing_state=ReleasingState.RELEASING
        )
        db.session.add(media_item)
        db.session.commit()
        return media_item

    def test_json_representation(self):
        """
        Tests the JSON representation of the model
        :return: None
        """
        media_item = self.generate_sample_media_item()

        self.assertEqual(
            media_item.__json__(False),
            {
                "id": media_item.id,
                "media_type": media_item.media_type.value,
                "media_subtype": media_item.media_subtype.value,
                "english_title": media_item.english_title,
                "romaji_title": media_item.romaji_title,
                "cover_url": media_item.cover_url,
                "latest_release": media_item.latest_release,
                "releasing_state": media_item.releasing_state.value
            }
        )
        self.assertEqual(
            media_item.__json__(True),
            media_item.__json__(False)
        )

    def test_string_representation(self):
        """
        Tests the string representation of the model
        :return: None
        """
        media_item = self.generate_sample_media_item()
        data = media_item.__json__()
        data.pop("id")
        self.assertEqual(
            str(media_item),
            "MediaItem:{} <{}>".format(media_item.id, data)
        )

    def test_repr(self):
        """
        Tests the __repr__ method of the model class
        :return: None
        """
        media_item = self.generate_sample_media_item()
        generated = {"value": media_item}
        code = repr(media_item)

        exec("generated['value'] = " + code)
        self.assertEqual(generated["value"], media_item)
        self.assertFalse(generated["value"] is media_item)

    def test_hashing(self):
        """
        Tests using the model objects as keys in a dictionary
        :return: None
        """
        media_item = self.generate_sample_media_item()
        media_item_2 = MediaItem(
            media_type=MediaType.MANGA,
            media_subtype=MediaSubType.MANGA,
            english_title="Don't Fly Me to the Moon",
            romaji_title="ATonikaku Cawaii",
            cover_url="https://s4.anilist.co/file/anilistcdn/media/manga/"
                      "cover/medium/nx101177-FjjD5UWB3C3t.png",
            latest_release=None,
            releasing_state=ReleasingState.RELEASING
        )
        db.session.add(media_item_2)
        db.session.commit()
        mapping = {
            media_item: 100,
            media_item_2: 200
        }
        self.assertEqual(mapping[media_item], 100)
        self.assertEqual(mapping[media_item_2], 200)

    def test_equality(self):
        """
        Tests checking equality for model objects
        :return: None
        """
        media_item = self.generate_sample_media_item()
        media_item_2 = MediaItem(
            media_type=MediaType.MANGA,
            media_subtype=MediaSubType.MANGA,
            english_title="Don't Fly Me to the Moon",
            romaji_title="ATonikaku Cawaii",
            cover_url="https://s4.anilist.co/file/anilistcdn/media/manga/"
                      "cover/medium/nx101177-FjjD5UWB3C3t.png",
            latest_release=None,
            releasing_state=ReleasingState.RELEASING
        )
        db.session.add(media_item_2)
        db.session.commit()
        self.assertEqual(media_item, media_item)
        self.assertNotEqual(media_item, media_item_2)
        self.assertNotEqual(media_item, 100)

    def test_title(self):
        """
        Tests the title attribute of the media item
        :return: None
        """
        media_item = self.generate_sample_media_item()
        self.assertEqual(media_item.title, "Fly Me to the Moon")
        media_item.english_title = None
        self.assertEqual(media_item.title, "Tonikaku Cawaii")

    def test_uniqueness(self):
        """
        Tests that a mediaitem can only exist once
        :return: None
        """
        standard_kwargs = {
            "media_type": MediaType.MANGA,
            "media_subtype": MediaSubType.MANGA,
            "english_title": "ABC",
            "romaji_title": "DEF",
            "cover_url": "N/A",
            "latest_release": None,
            "releasing_state": ReleasingState.RELEASING
        }
        original = MediaItem(**standard_kwargs)
        db.session.add(original)
        db.session.commit()

        for key, value, error_expected in [
            ("media_type", MediaType.ANIME, False),
            ("media_subtype", MediaSubType.NOVEL, False),
            ("romaji_title", "XYZ", False),
            ("cover_url", "AAA", False),
            ("english_title", "ABCDE", True),
        ]:
            kwargs = dict(standard_kwargs)
            kwargs[key] = value
            try:
                generated = MediaItem(**kwargs)
                db.session.add(generated)
                db.session.commit()
                if error_expected:
                    self.fail()
            except IntegrityError as e:
                db.session.rollback()
                if not error_expected:
                    raise e



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Optional, Dict, Any, Tuple
from otaku_info.utils.enums import MediaType, MediaSubType, \
    ReleasingState, ConsumingState, MediaRelationType


class AnilistItem:
    """
    Class that models a general anilist list item
    Represents the information fetched using anilist's API
    """
    def __init__(
            self,
            anilist_id: int,
            media_type: MediaType,
            media_subtype: MediaSubType,
            english_title: Optional[str],
            romaji_title: str,
            cover_url: str,
            chapters: Optional[int],
            episodes: Optional[int],
            releasing_state: ReleasingState,
            relations: Dict[Tuple[MediaType, int], MediaRelationType]
    ):
        """
        Initializes the AnilistItem object
        :param anilist_id: The anilist ID of the series
        :param media_type: The media type of the series
        :param media_subtype: The media subtype of the series
        :param english_title: The English title of the series
        :param romaji_title: The Japanes title of the series written in romaji
        :param cover_url: URL to a cover image for the series
        :param chapters: The total amount of known manga chapters
        :param episodes: The total amount of known anime episodes
        :param releasing_state: The current releasing state of the series
        :param relations: Related media items identified by IDs
        """
        self.anilist_id = anilist_id
        self.media_type = media_type
        self.media_subtype = media_subtype
        self.english_title = english_title
        self.romaji_title = romaji_title
        self.cover_url = cover_url
        self.chapters = chapters
        self.episodes = episodes
        self.releasing_state = releasing_state
        self.relations = relations

    @property
    def latest_release(self) -> Optional[int]:
        """
        :return: The latest release. Chapters for manga, episodes for anime
        """
        if self.media_type == MediaType.ANIME:
            return self.episodes
        else:
            return self.chapters

    @classmethod
    def from_query(cls, media_type: MediaType, data: Dict[str, Any]) \
            -> "AnilistItem":
        """
        Generates an AnilistItem from a dictionary generated by an APi query
        :param media_type: The media type of the item
        :param data: The data to use
        :return: The generated AnilistItem
        """
        _media_subtype = data["format"]
        if _media_subtype is None:
            _media_subtype = "unknown"
        media_subtype = MediaSubType(_media_subtype.lower())

        _releasing_state = data["status"]
        if _releasing_state is None:
            _releasing_state = "unknown"
        releasing_state = ReleasingState(_releasing_state.lower())

        relations = {}
        for edge in data["relations"]["edges"]:
            node_id = edge["node"]["id"]
            node_type = MediaType(edge["node"]["type"].lower())
            relation_type = MediaRelationType(edge["relationType"].lower())
            relations[(node_type, node_id)] = relation_type

        return AnilistItem(
            data["id"],
            media_type,
            media_subtype,
            data["title"]["english"],
            data["title"]["romaji"],
            data["coverImage"]["large"],
            data["chapters"],
            data["episodes"],
            releasing_state,
            relations
        )


class AnilistUserItem(AnilistItem):
    """
    Class that models an anilist list item for a user
    Represents the information fetched using anilist's API
    """
    def __init__(
            self,
            anilist_id: int,
            media_type: MediaType,
            media_subtype: MediaSubType,
            english_title: Optional[str],
            romaji_title: str,
            cover_url: str,
            chapters: Optional[int],
            episodes: Optional[int],
            releasing_state: ReleasingState,
            relations: Dict[Tuple[MediaType, int], MediaRelationType],
            score: Optional[int],
            progress: Optional[int],
            consuming_state: ConsumingState,
            list_name: str
    ):
        """
        Initializes the AnilistItem object
        :param anilist_id: The anilist ID of the series
        :param media_type: The media type of the series
        :param media_subtype: The media subtype of the series
        :param english_title: The English title of the series
        :param romaji_title: The Japanes title of the series written in romaji
        :param cover_url: URL to a cover image for the series
        :param chapters: The total amount of known manga chapters
        :param episodes: The total amount of known anime episodes
        :param releasing_state: The current releasing state of the series
        :param relations: Related media items identified by IDs
        :param score: The user's score for the series
        :param progress: The user's progress for the series
        :param consuming_state: The user's consumption state for the series
        :param list_name: Which of the user's lists this entry belongs to
        """
        super().__init__(
            anilist_id,
            media_type,
            media_subtype,
            english_title,
            romaji_title,
            cover_url,
            chapters,
            episodes,
            releasing_state,
            relations
        )
        self.score = score
        self.progress = progress
        self.consuming_state = consuming_state
        self.list_name = list_name

    @classmethod
    def from_query(
            cls,
            media_type: MediaType,
            data: Dict[str, Any]
    ) -> "AnilistUserItem":
        """
        Generates an AnilistUserItem from a dictionary generated
        by an APi query
        :param media_type: The media type of the item
        :param data: The data to use
        :return: The generated AnilistItem
        """
        base = AnilistItem.from_query(media_type, data["media"])
        consuming_state = ConsumingState(data["status"].lower())

        return AnilistUserItem(
            base.anilist_id,
            base.media_type,
            base.media_subtype,
            base.english_title,
            base.romaji_title,
            base.cover_url,
            base.chapters,
            base.episodes,
            base.releasing_state,
            base.relations,
            data["score"],
            data["progress"],
            consuming_state,
            data["list_name"]
        )



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from flask import request, redirect, url_for
from flask.blueprints import Blueprint
from flask_login import login_required, current_user
from puffotter.flask.base import db
from otaku_info.utils.enums import ListService
from otaku_info.db.ServiceUsername import ServiceUsername


def define_blueprint(blueprint_name: str) -> Blueprint:
    """
    Defines the blueprint for this route
    :param blueprint_name: The name of the blueprint
    :return: The blueprint
    """
    blueprint = Blueprint(blueprint_name, __name__)

    @blueprint.route("/set_service_username", methods=["POST"])
    @login_required
    def set_service_username():
        """
        Sets an exernal service username for the current user
        :return:
        """
        username = request.form["username"]
        service = ListService(request.form["service"])

        service_username: ServiceUsername = ServiceUsername.query\
            .filter_by(user=current_user, service=service).first()
        if service_username is None:
            service_username = ServiceUsername(
                user=current_user, username=username, service=service
            )
            db.session.add(service_username)
        else:
            service_username.username = username

        db.session.commit()
        return redirect(url_for("user_management.profile"))

    return blueprint



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from datetime import datetime
from typing import Dict, Any, Optional, List
from puffotter.flask.base import db
from puffotter.flask.db.ModelMixin import ModelMixin
from otaku_info.db.MediaItem import MediaItem
from otaku_info.db.MediaId import MediaId


class LnRelease(ModelMixin, db.Model):
    """
    Database model that keeps track of light novel releases
    """

    __tablename__ = "ln_releases"
    """
    The name of the database table
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes the Model
        :param args: The constructor arguments
        :param kwargs: The constructor keyword arguments
        """
        super().__init__(*args, **kwargs)

    media_item_id: Optional[int] = db.Column(
        db.Integer,
        db.ForeignKey(
            "media_items.id", ondelete="CASCADE", onupdate="CASCADE"
        ),
        nullable=True,
        unique=False
    )
    """
    The ID of the media item referenced by this light novel release
    """

    media_item: Optional[MediaItem] = db.relationship(
        "MediaItem",
        backref=db.backref(
            "ln_releases", lazy=True, cascade="all,delete"
        )
    )
    """
    The media item referenced by this light novel release
    """

    release_date_string: str = db.Column(db.String(10), nullable=False)
    """
    The release date as a ISO-8601 string
    """

    series_name: str = db.Column(db.String(255), nullable=False)
    """
    The series name
    """

    volume: str = db.Column(db.String(255), nullable=False)
    """
    The volume identifier
    """

    publisher: Optional[str] = db.Column(db.String(255), nullable=True)
    """
    The publisher
    """

    purchase_link: Optional[str] = db.Column(db.String(255), nullable=True)
    """
    Link to a store page
    """

    digital: bool = db.Column(db.Boolean)
    """
    Whether this is a digital release
    """

    physical: bool = db.Column(db.Boolean)
    """
    Whether this is a physical release
    """

    @property
    def release_date(self) -> datetime:
        """
        :return: The release date as a datetime object
        """
        return datetime.strptime(self.release_date_string, "%Y-%m-%d")

    def __json__(self, include_children: bool = False) -> Dict[str, Any]:
        """
        Generates a dictionary containing the information of this model
        :param include_children: Specifies if children data models
                                 will be included or if they're limited to IDs
        :return: A dictionary representing the model's values
        """
        data = {
            "id": self.id,
            "media_item_id": self.media_item_id,
            "release_date_string": self.release_date_string,
            "series_name": self.series_name,
            "volume": self.volume,
            "publisher": self.publisher,
            "purchase_link": self.purchase_link,
            "digital": self.digital,
            "physical": self.physical
        }
        if include_children:
            if self.media_item is None:
                data["media_item"] = None
            else:
                data["media_item"] = self.media_item.__json__(include_children)
        return data

    def get_ids(self) -> List[MediaId]:
        """
        :return: Any related Media IDs
        """
        if self.media_item is None:
            return []
        else:
            return MediaId.query.filter_by(media_item=self.media_item).all()



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from otaku_info.test.TestFramework import _TestFramework
from otaku_info.utils.anilist.api import load_anilist, \
    guess_latest_manga_chapter, load_media_info
from otaku_info.utils.enums import MediaType


class TestApi(_TestFramework):
    """
    Class that tests the anilist API functions
    """

    def test_loading_anilist(self):
        """
        Tests loading a user's anilist
        :return: None
        """
        load_anilist("namboy94", MediaType.MANGA)
        load_anilist("namboy94", MediaType.ANIME)

    def test_guessing_manga_chapter(self):
        """
        Tests guessing a manga chapter
        :return: None
        """
        guess = guess_latest_manga_chapter(101177)
        self.assertGreater(guess, 100)

    def test_loading_media(self):
        """
        Tests loading a media item
        :return: None
        """
        media = load_media_info(1, MediaType.ANIME)
        self.assertEqual(media.english_title, "Cowboy Bebop")
        self.assertGreater(len(media.relations), 0)



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import requests
from typing import List, Optional
from datetime import datetime
from bs4 import BeautifulSoup
from puffotter.flask.base import db
from otaku_info.db.MediaId import MediaId
from otaku_info.db.MediaItem import MediaItem
from otaku_info.db.LnRelease import LnRelease
from otaku_info.utils.dates import map_month_name_to_month_number
from otaku_info.utils.enums import MediaType, ListService
from otaku_info.utils.anilist.api import load_media_info, \
    map_myanimelist_id_to_anilist_id
from otaku_info.background.anilist import update_media_id, update_media_item


def load_tables(year: int) -> List[BeautifulSoup]:
    """
    Loads the tables containing the release data
    """
    current_year = datetime.utcnow().year

    # TODO Parse years from 2015-2017
    if year < 2018 or year > current_year + 1:
        return []

    if year >= current_year:
        url = "https://old.reddit.com/r/LightNovels/wiki/upcomingreleases"
    else:
        url = f"https://old.reddit.com/r/LightNovels/wiki/{year}releases"

    headers = {"User-Agent": "Mozilla/5.0"}
    resp = requests.get(url, headers=headers).text
    soup = BeautifulSoup(resp, "html.parser")

    tables = soup.find_all("tbody")

    # Table 0: Releases for current month on side bar
    # Table 1: Table below current month releases on side bar
    # Table -1: To be announced
    tables = tables[2:-1]

    if year >= current_year:
        if year == current_year:
            tables = tables[0:12]
        else:
            tables = tables[12:]

    return tables


def parse_release(year: int, parts: List[BeautifulSoup]) -> LnRelease:
    """
    Parses a light novel release table entry and turns it into
    an LnRelease object.
    The LnRelease entry in the database is also updated if it already exists
    :param year: The year of the release
    :param parts: The parts of the table entry
    :return: The resulting LnRelease object
    """

    release_date_string = parts[0].text
    month_name, day_string = release_date_string.split(" ")
    month = map_month_name_to_month_number(month_name)
    if month is None:
        month = 1

    try:
        day = int(day_string)
    except ValueError:
        day = 1
    release_date = datetime(year=year, month=month, day=day)
    release_date_string = release_date.strftime("%Y-%m-%d")

    series_name = parts[1].text
    volume = parts[2].text
    publisher = parts[3].text

    purchase_link_item = parts[3].find("a")
    purchase_link = None
    if purchase_link_item is not None:
        purchase_link = purchase_link_item["href"]

    info_link_item = parts[1].find("a")
    info_link: Optional[str] = None
    if info_link_item is not None:
        info_link = info_link_item["href"]

    digital = "digital" in parts[4].text.lower()
    physical = "physical" in parts[4].text.lower()

    media_item = get_media_item(info_link)

    release: LnRelease = LnRelease.query.filter_by(
        series_name=series_name,
        volume=volume,
        digital=digital,
        physical=physical,
        publisher=publisher
    ).first()
    is_new = False
    if release is None:
        release = LnRelease(
            series_name=series_name,
            volume=volume,
            digital=digital,
            physical=physical,
            publisher=publisher
        )
        is_new = True

    release.release_date_string = release_date_string
    release.purchase_link = purchase_link
    release.media_item = media_item

    if is_new:
        db.session.add(release)

    db.session.commit()
    return release


def get_media_item(info_link: Optional[str]) -> Optional[MediaItem]:
    """
    Retrieves the media item for an info link
    Automatically adds missing entries to the database
    :param info_link: The info link
    :return: The media item or None if no media item can be found
    """

    if info_link is None or "myanimelist.net" not in info_link:
        return None

    url_parts = info_link.split("/")
    index = -1
    while not url_parts[index].isdigit():
        index -= 1
    myanimelist_id = url_parts[index]

    existing = MediaId.query.filter_by(
        service_id=myanimelist_id,
        service=ListService.MYANIMELIST
    ).all()
    existing = [
        x for x in existing
        if x.media_item.media_type == MediaType.MANGA
    ]

    if len(existing) != 0:
        return existing[0].media_item
    else:
        anilist_id = map_myanimelist_id_to_anilist_id(
            int(myanimelist_id), MediaType.MANGA
        )

        if anilist_id is None:
            return None

        existing = MediaId.query.filter_by(
            service_id=str(anilist_id),
            service=ListService.ANILIST
        ).all()
        existing = [
            x for x in existing
            if x.media_item.media_type == MediaType.MANGA
        ]

        if len(existing) == 0:
            media_info = load_media_info(anilist_id, MediaType.MANGA)
            if media_info is None:
                media_item = None
            else:
                media_item = update_media_item(media_info, None)
                update_media_id(media_info, media_item, None)
        else:
            anilist_id_obj = existing[0]
            media_item = anilist_id_obj.media_item

        if media_item is not None:
            myanimelist_id_obj = MediaId(
                media_item=media_item,
                service=ListService.MYANIMELIST,
                service_id=str(myanimelist_id)
            )
            db.session.add(myanimelist_id_obj)
            db.session.commit()
        return media_item


def load_ln_releases(year: int) -> List[LnRelease]:
    """
    Loads the currently available light novel releases from reddit's
    /r/lightnovels subreddit.
    :param year: Specifies the year of the releases
    :return: The releases as LnRelease objects
    """
    tables = load_tables(year)

    releases: List[LnRelease] = []
    for i, table in enumerate(tables):
        month_number = i + 1

        for entry in table.find_all("tr"):
            release = parse_release(year, entry.find_all("td"))

            if month_number != release.release_date.month:
                print(f"Incorrect month: "
                      f"{month_number} != {release.release_date.month} "
                      f"({release.release_date_string})")

            releases.append(release)

    return releases



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Optional, List
from puffotter.graphql import GraphQlClient
from otaku_info.utils.enums import MediaType
from otaku_info.utils.anilist.AnilistItem \
    import AnilistItem, AnilistUserItem


MEDIA_QUERY = """
    id
    chapters
    episodes
    status
    format
    title {
        english
        romaji
    }
    coverImage {
        large
    }
    nextAiringEpisode {
      episode
    }
    relations {
        edges {
            node {
                id
                type
            }
            relationType
        }
    }
"""


def guess_latest_manga_chapter(anilist_id: int) -> Optional[int]:
    """
    Guesses the latest chapter number based on anilist user activity
    :param anilist_id: The anilist ID to check
    :return: The latest chapter number
    """
    graphql = GraphQlClient("https://graphql.anilist.co")
    query = """
    query ($id: Int) {
        Page(page: 1) {
        activities(mediaId: $id, sort: ID_DESC) {
                ... on ListActivity {
                    progress
                    userId
                    status
                    media {
                        chapters
                    }
                }
            }
        }
    }
    """
    resp = graphql.query(query, {"id": anilist_id})
    if resp is None:
        return None

    data = resp["data"]["Page"]["activities"]

    progresses = []
    for entry in data:
        progress = entry["progress"]
        status = entry["status"]
        chapters = entry["media"]["chapters"]

        if status == "completed":
            progresses.append(chapters)
        elif progress is not None:
            progress = entry["progress"].split(" - ")[-1]
            progresses.append(int(progress))

    progresses = progresses[0:20]
    progresses.sort(key=lambda x: progresses.count(x), reverse=True)
    progresses = sorted(progresses, key=progresses.count, reverse=True)

    try:
        return progresses[0]
    except IndexError:
        return None


def load_anilist(
        username: str,
        media_type: MediaType
) -> List[AnilistUserItem]:
    """
    Loads the anilist for a user
    :param username: The username
    :param media_type: The media type, either MANGA or ANIME
    :return: The anilist list items for the user and media type
    """
    graphql = GraphQlClient("https://graphql.anilist.co")
    query = """
    query ($username: String, $media_type: MediaType) {
        MediaListCollection(userName: $username, type: $media_type) {
            lists {
                name
                entries {
                    progress
                    score
                    status
                    media {@{MEDIA_QUERY}}
                }
            }
        }
    }
    """.replace("@{MEDIA_QUERY}", MEDIA_QUERY)
    resp = graphql.query(query, {
        "username": username,
        "media_type": media_type.value.upper()
    })
    if resp is None:
        return []
    user_lists = resp["data"]["MediaListCollection"]["lists"]

    anilist_items: List[AnilistUserItem] = []
    for entries, list_name in [
        (y["entries"], y["name"]) for y in user_lists
    ]:
        for entry in entries:
            entry["list_name"] = list_name
            anilist_items.append(AnilistUserItem.from_query(media_type, entry))

    return anilist_items


def load_media_info(anilist_id: int, media_type: MediaType) \
        -> Optional[AnilistItem]:
    """
    Loads information for a single anilist media item
    :param anilist_id: The anilist media ID
    :param media_type: The media type
    :return: The fetched AnilistItem
    """
    graphql = GraphQlClient("https://graphql.anilist.co")
    query = """
        query ($id: Int, $media_type: MediaType) {
            Media(id: $id, type: $media_type) {
                @{MEDIA_QUERY}
            }
        }
    """.replace("@{MEDIA_QUERY}", MEDIA_QUERY)
    resp = graphql.query(
        query,
        {"id": anilist_id, "media_type": media_type.value.upper()}
    )
    if resp is None:
        return None
    else:
        return AnilistItem.from_query(media_type, resp["data"]["Media"])


def map_myanimelist_id_to_anilist_id(
        myanimelist_id: int,
        media_type: MediaType
) -> Optional[int]:
    """
    Translates a myanimelist ID to an Anilist ID
    :param myanimelist_id: The myanimelist ID
    :param media_type: The media type
    :return: The anilist ID
    """
    graphql = GraphQlClient("https://graphql.anilist.co")
    query = """
        query ($mal_id: Int, $type: MediaType) {
            Media(idMal: $mal_id, type: $type) {
                id
            }
        }
    """
    resp = graphql.query(
        query,
        {"mal_id": myanimelist_id, "media_type": media_type.value.upper()}
    )
    if resp is None:
        return None
    else:
        return int(resp["data"]["Media"]["id"])



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import List, Tuple, Dict, Any
from puffotter.flask.db.User import User
from puffotter.flask.base import app
from otaku_info.db.MangaChapterGuess import MangaChapterGuess
from otaku_info.db.MediaItem import MediaItem
from otaku_info.db.MediaList import MediaList
from otaku_info.db.MediaListItem import MediaListItem
from otaku_info.db.MediaUserState import MediaUserState
from otaku_info.db.MediaId import MediaId
from otaku_info.utils.manga_updates.MangaUpdate import MangaUpdate
from otaku_info.utils.db_model_helper import build_title
from otaku_info.utils.enums \
    import MediaType, MediaSubType, ConsumingState, ReleasingState, ListService


def load_applicable_data(
        user: User,
        service: ListService,
        media_list: str,
        include_complete: bool
) -> Tuple[
    Dict[int, Dict[str, Any]],
    Dict[int, Dict[str, Any]],
    Dict[int, Dict[str, Any]],
    Dict[int, Dict[str, Any]],
    Dict[int, Dict[str, Any]],
    Dict[int, int]
]:
    """
    Loads the applicable data from the database in an efficient manner.
    By only loading the data we need and avoiding JOINs, the performance
    is increased drastically.
    Since this method is called for every call to a manga/updates page,
    this should be fast.
    The return values are mostly database IDs mapped to dictionaries
    containing the data required for displaying manga updates.
    A notable exception are the manga chapter guesses, which are simply
    MediaId IDs mapped to the chapter guess value.
    :param user: The user requesting the manga updates
    :param service: The service for which to fetch the updates
    :param media_list: The media list for which to fetch the updates
    :param include_complete: Whether or not to include completed series
    :return: media items, media ids, media user states, media lists,
             media list items, manga chapter guesses
    """

    applicable_releasing_states = [ReleasingState.RELEASING]
    if include_complete:
        applicable_releasing_states += [
            ReleasingState.FINISHED,
            ReleasingState.CANCELLED
        ]

    media_items: Dict[int, Dict[str, Any]] = {
        x[0]: {
            "title": build_title(x[1], x[2]),
            "cover_url": x[3],
            "latest_release": x[4]
        }
        for x in MediaItem.query
        .with_entities(
            MediaItem.id,
            MediaItem.english_title,
            MediaItem.romaji_title,
            MediaItem.cover_url,
            MediaItem.latest_release,
            MediaItem.releasing_state
        )
        .filter(MediaItem.media_type == MediaType.MANGA)
        .filter(MediaItem.media_subtype == MediaSubType.MANGA)
        .all()
        if x[5] in applicable_releasing_states
    }

    media_ids: Dict[int, Dict[str, Any]] = {
        x[0]: {
            "media_item_id": x[1],
            "service": x[2],
            "service_id": x[3]
        }
        for x in MediaId.query
        .with_entities(
            MediaId.id,
            MediaId.media_item_id,
            MediaId.service,
            MediaId.service_id
        )
        .all()
        if x[1] in media_items
    }

    user_states: Dict[int, Dict[str, Any]] = {
        x[0]: {
            "media_id_id": x[1],
            "progress": x[2],
            "score": x[3]
        }
        for x in MediaUserState.query
        .with_entities(
            MediaUserState.id,
            MediaUserState.media_id_id,
            MediaUserState.progress,
            MediaUserState.score,
            MediaUserState.consuming_state
        )
        .all()
        if x[1] in media_ids and x[4] in [
            ConsumingState.CURRENT, ConsumingState.PAUSED
        ]
    }

    user_lists: Dict[int, Dict[str, Any]] = {
        x[0]: {}
        for x in MediaList.query
        .with_entities(MediaList.id)
        .filter_by(user_id=user.id)
        .filter_by(name=media_list)
        .filter_by(service=service)
        .all()
    }

    list_items: Dict[int, Dict[str, Any]] = {
        x[0]: {
            "media_list_id": x[1],
            "media_user_state_id": x[2]
        }
        for x in MediaListItem.query
        .with_entities(
            MediaListItem.id,
            MediaListItem.media_list_id,
            MediaListItem.media_user_state_id
        )
        .all()
        if x[1] in user_lists and x[2] in user_states
    }

    chapter_guesses: Dict[int, int] = {
        x.media_id_id: x.guess
        for x in MangaChapterGuess.query.all()
    }

    return media_items, media_ids, user_states, user_lists, list_items, \
        chapter_guesses


def prepare_manga_updates(
        user: User,
        service: ListService,
        media_list: str,
        include_complete: bool,
        min_update_count: int
) -> List[MangaUpdate]:
    """
    Prepares easily understandable objects to display for manga updates
    :param user: The user requesting the manga updates
    :param service: The service for which to fetch the updates
    :param media_list: The media list for which to fetch the updates
    :param include_complete: Whether or not to include completed series
    :param min_update_count: The minimum amount of new chapters required
                             for an update to be generated
    :return: A list of MangaUpdate objects, sorted by score
    """
    app.logger.debug("Starting preparing manga updates")

    import time
    start = time.time()

    media_items, media_ids, user_states, user_lists, list_items, \
        chapter_guesses = load_applicable_data(
            user, service, media_list, include_complete
        )

    media_item_service_ids: Dict[int, List[Tuple[ListService, str]]] = {
        x: [] for x in media_items
    }
    for _, media_id in media_ids.items():
        media_item_service_ids[media_id["media_item_id"]].append(
            (media_id["service"], media_id["service_id"])
        )

    combined = []
    for media_list_item_id, media_list_item in list_items.items():
        data = media_list_item
        data.update(user_lists[data["media_list_id"]])
        data.update(user_states[data["media_user_state_id"]])
        data.update(media_ids[data["media_id_id"]])
        data.update(media_items[data["media_item_id"]])
        data["guess"] = chapter_guesses.get(data["media_id_id"])

        data["related_ids"] = media_item_service_ids[data["media_item_id"]]

        combined.append(data)

    compiled = [
        MangaUpdate(
            x["media_item_id"],
            x["title"],
            x["cover_url"],
            x["latest_release"],
            x["progress"],
            x["score"],
            x["guess"],
            x["related_ids"]
        )
        for x in combined
    ]

    app.logger.warning(time.time() - start)

    return list(filter(lambda x: x.diff >= min_update_count, compiled))



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Tuple
from puffotter.flask.base import db
from puffotter.flask.db.User import User
from sqlalchemy.exc import IntegrityError
from otaku_info.db.MediaItem import MediaItem
from otaku_info.db.MediaId import MediaId
from otaku_info.db.MediaUserState import MediaUserState
from otaku_info.db.MediaList import MediaList
from otaku_info.db.MediaListItem import MediaListItem
from otaku_info.utils.enums import ListService, MediaType, MediaSubType, \
    ReleasingState, ConsumingState
from otaku_info.test.TestFramework import _TestFramework


class TestMediaListItem(_TestFramework):
    """
    Class that tests the MediaListItem database model
    """

    def generate_sample_media_list_item(self) -> Tuple[
        MediaListItem, MediaList, MediaUserState, User, MediaItem, MediaId
    ]:
        """
        Generates a media list item
        :return: The media list item, media list, media user state, user,
                 media item and media id
        """
        user = self.generate_sample_user(True)[0]
        media_item = MediaItem(
            media_type=MediaType.MANGA,
            media_subtype=MediaSubType.MANGA,
            english_title="Fly Me to the Moon",
            romaji_title="Tonikaku Cawaii",
            cover_url="https://s4.anilist.co/file/anilistcdn/media/manga/"
                      "cover/medium/nx101177-FjjD5UWB3C3t.png",
            latest_release=None,
            releasing_state=ReleasingState.RELEASING
        )
        media_id = MediaId(
            media_item=media_item,
            service_id="101177",
            service=ListService.ANILIST
        )
        media_user_state = MediaUserState(
            media_id=media_id,
            user=user,
            progress=10,
            score=75,
            consuming_state=ConsumingState.CURRENT
        )
        media_list = MediaList(
            name="ABC",
            user=user,
            service=ListService.ANILIST,
            media_type=MediaType.MANGA
        )
        media_list_item = MediaListItem(
            media_list=media_list,
            media_user_state=media_user_state
        )
        db.session.add(media_item)
        db.session.add(media_id)
        db.session.add(media_user_state)
        db.session.add(media_list)
        db.session.add(media_list_item)
        db.session.commit()
        return media_list_item, media_list, media_user_state, \
            user, media_item, media_id

    def test_json_representation(self):
        """
        Tests the JSON representation of the model
        :return: None
        """
        media_list_item, media_list, media_user_state, \
            user, media_item, media_id = \
            self.generate_sample_media_list_item()

        self.assertEqual(
            media_list_item.__json__(False),
            {
                "id": media_id.id,
                "media_list_id": media_list.id,
                "media_user_state_id": media_user_state.id
            }
        )
        self.assertEqual(
            media_list_item.__json__(True),
            {
                "id": media_id.id,
                "media_list_id": media_list.id,
                "media_list": media_list.__json__(True),
                "media_user_state_id": media_user_state.id,
                "media_user_state": media_user_state.__json__(True)
            }
        )

    def test_string_representation(self):
        """
        Tests the string representation of the model
        :return: None
        """
        media_list_item, media_list, media_user_state, \
            user, media_item, media_id = \
            self.generate_sample_media_list_item()
        data = media_list_item.__json__()
        data.pop("id")
        self.assertEqual(
            str(media_list_item),
            "MediaListItem:{} <{}>".format(media_list_item.id, data)
        )

    def test_repr(self):
        """
        Tests the __repr__ method of the model class
        :return: None
        """
        media_list_item, media_list, media_user_state, \
            user, media_item, media_id = \
            self.generate_sample_media_list_item()
        generated = {"value": media_list_item}
        code = repr(media_list_item)

        exec("generated['value'] = " + code)
        self.assertEqual(generated["value"], media_list_item)
        self.assertFalse(generated["value"] is media_list_item)

    def test_hashing(self):
        """
        Tests using the model objects as keys in a dictionary
        :return: None
        """
        media_list_item, media_list, media_user_state, \
            user, media_item, media_id = \
            self.generate_sample_media_list_item()
        media_list_2 = MediaList(
            name="XYZ",
            user=user,
            service=ListService.ANILIST,
            media_type=MediaType.MANGA
        )
        media_list_item_2 = MediaListItem(
            media_list=media_list_2,
            media_user_state=media_user_state
        )
        db.session.add(media_list_2)
        db.session.add(media_list_item_2)
        db.session.commit()
        mapping = {
            media_list_item: 100,
            media_list_item_2: 200
        }
        self.assertEqual(mapping[media_list_item], 100)
        self.assertEqual(mapping[media_list_item_2], 200)

    def test_equality(self):
        """
        Tests checking equality for model objects
        :return: None
        """
        media_list_item, media_list, media_user_state, \
            user, media_item, media_id = \
            self.generate_sample_media_list_item()
        media_list_2 = MediaList(
            name="XYZ",
            user=user,
            service=ListService.ANILIST,
            media_type=MediaType.MANGA
        )
        media_list_item_2 = MediaListItem(
            media_list=media_list_2,
            media_user_state=media_user_state
        )
        db.session.add(media_list_2)
        db.session.add(media_list_item_2)
        db.session.commit()
        self.assertEqual(media_list_item, media_list_item)
        self.assertNotEqual(media_list_item, media_list_item_2)
        self.assertNotEqual(media_list_item, 100)

    def test_uniqueness(self):
        """
        Tests if the uniqueness of the model is handled properly
        :return: None
        """
        media_list_item, media_list, media_user_state, \
            user, media_item, media_id = \
            self.generate_sample_media_list_item()

        standard_kwargs = media_list_item.__json__(False)
        standard_kwargs.pop("id")

        media_list_kwargs = media_list.__json__(False)
        media_list_kwargs.pop("id")
        media_list_kwargs["service"] = ListService.KITSU
        media_list_kwargs["media_type"] = MediaType.ANIME
        new_media_list = MediaList(**media_list_kwargs)
        db.session.add(new_media_list)
        db.session.commit()

        media_user_state_kwargs = media_user_state.__json__(False)
        media_user_state_kwargs.pop("id")
        media_user_state_kwargs["user_id"] = \
            self.generate_sample_user(True)[0].id
        media_user_state_kwargs["consuming_state"] \
            = media_user_state.consuming_state
        new_media_user_state = MediaUserState(**media_user_state_kwargs)
        db.session.add(new_media_user_state)
        db.session.commit()

        try:
            duplicate = MediaListItem(**standard_kwargs)
            db.session.add(duplicate)
            db.session.commit()
            self.fail()
        except IntegrityError:
            db.session.rollback()

        for key, value in [
            ("media_list_id", new_media_list.id),
            ("media_user_state_id", new_media_user_state.id)
        ]:
            kwargs = dict(standard_kwargs)
            kwargs[key] = value

            generated = MediaListItem(**kwargs)
            db.session.add(generated)
            db.session.commit()



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from enum import Enum


class MediaType(Enum):
    """
    Class that models a media type for media items
    """
    ANIME = "anime"
    MANGA = "manga"


class MediaSubType(Enum):
    """
    Class that models a media subtype for media items
    """
    TV = "tv"
    TV_SHORT = "tv_short"
    MOVIE = "movie"
    SPECIAL = "special"
    OVA = "ova"
    ONA = "ona"
    MUSIC = "music"
    MANGA = "manga"
    NOVEL = "novel"
    ONE_SHOT = "one_shot"
    UNKNOWN = "unknown"


class MediaRelationType(Enum):
    """
    Class that models a media relation type
    """
    ADAPTATION = "adaptation"
    PREQUEL = "prequel"
    SEQUEL = "sequel"
    PARENT = "parent"
    SIDE_STORY = "side_story"
    CHARACTER = "character"
    SUMMARY = "summary"
    ALTERNATIVE = "alternative"
    SPIN_OFF = "spin_off"
    OTHER = "other"
    SOURCE = "source"
    COMPILATION = "compilation"
    CONTAINS = "contains"


class ListService(Enum):
    """
    Class that defines available list services
    """
    ANILIST = "anilist"
    MYANIMELIST = "myanimelist"
    MANGADEX = "mangadex"
    MANGAUPDATES = "mangaupdates"
    KITSU = "kitsu"
    ANIMEPLANET = "animeplanet"


class ReleasingState(Enum):
    """
    Class that defines possible releasing states
    """
    FINISHED = "finished"
    RELEASING = "releasing"
    NOT_YET_RELEASED = "not_yet_released"
    CANCELLED = "cancelled"
    UNKNOWN = "unknown"


class ConsumingState(Enum):
    """
    Class that defines the possible consuming states for a user and media item
    """
    CURRENT = "current"
    PLANNING = "planning"
    COMPLETED = "completed"
    DROPPED = "dropped"
    PAUSED = "paused"
    REPEATING = "repeating"


class NotificationType(Enum):
    """
    Class that defines the possible notification types
    """
    NEW_MANGA_CHAPTERS = "new_manga_chapters"
    NEW_ANIME_EPISODES = "new_anime_episodes"



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Dict, Any
from puffotter.flask.base import db
from puffotter.flask.db.ModelMixin import ModelMixin
from otaku_info.db.MediaList import MediaList
from otaku_info.db.MediaUserState import MediaUserState


class MediaListItem(ModelMixin, db.Model):
    """
    Database model for media list items.
    This model maps MediaLists and MediaUserStates
    """

    __tablename__ = "media_list_items"
    """
    The name of the database table
    """

    __table_args__ = (
        db.UniqueConstraint(
            "media_list_id",
            "media_user_state_id",
            name="unique_media_list_item"
        ),
    )
    """
    Makes sure that objects that should be unique are unique
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes the Model
        :param args: The constructor arguments
        :param kwargs: The constructor keyword arguments
        """
        super().__init__(*args, **kwargs)

    media_list_id: int = db.Column(
        db.Integer,
        db.ForeignKey(
            "media_lists.id", ondelete="CASCADE", onupdate="CASCADE"
        ),
        nullable=False
    )
    """
    The ID of the media list this list item is a part of
    """

    media_list: MediaList = db.relationship(
        "MediaList",
        backref=db.backref("media_list_items", lazy=True, cascade="all,delete")
    )
    """
    The media list this list item is a part of
    """

    media_user_state_id: int = db.Column(
        db.Integer,
        db.ForeignKey(
            "media_user_states.id", ondelete="CASCADE", onupdate="CASCADE"
        ),
        nullable=False
    )
    """
    The ID of the media user state this list item references
    """

    media_user_state: MediaUserState = db.relationship(
        "MediaUserState",
        backref=db.backref("media_list_items", lazy=True, cascade="all,delete")
    )
    """
    The media user state this list item references
    """

    def __json__(self, include_children: bool = False) -> Dict[str, Any]:
        """
        Generates a dictionary containing the information of this model
        :param include_children: Specifies if children data models
                                 will be included or if they're limited to IDs
        :return: A dictionary representing the model's values
        """
        data = {
            "id": self.id,
            "media_list_id": self.media_list_id,
            "media_user_state_id": self.media_user_state_id
        }
        if include_children:
            data["media_list"] = self.media_list.__json__(include_children)
            data["media_user_state"] = \
                self.media_user_state.__json__(include_children)
        return data



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from flask import render_template, abort
from flask.blueprints import Blueprint
from otaku_info.db.MediaItem import MediaItem
from otaku_info.db.MediaId import MediaId


def define_blueprint(blueprint_name: str) -> Blueprint:
    """
    Defines the blueprint for this route
    :param blueprint_name: The name of the blueprint
    :return: The blueprint
    """
    blueprint = Blueprint(blueprint_name, __name__)

    @blueprint.route("/media/<media_item_id>", methods=["GET"])
    def media(media_item_id: int):
        media_item = MediaItem.query.get(media_item_id)

        if media_item is None:
            abort(404)

        media_ids = MediaId.query.filter_by(media_item_id=media_item_id).all()
        return render_template(
            "media/media.html",
            media_item=media_item,
            media_ids=media_ids
        )

    return blueprint



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Dict, Any
from puffotter.flask.base import db
from puffotter.flask.db.User import User
from puffotter.flask.db.ModelMixin import ModelMixin
from otaku_info.utils.enums import NotificationType


class NotificationSetting(ModelMixin, db.Model):
    """
    Database model that stores notification settings for a user
    """

    __tablename__ = "notification_settings"
    """
    The name of the database table
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes the Model
        :param args: The constructor arguments
        :param kwargs: The constructor keyword arguments
        """
        super().__init__(*args, **kwargs)

    user_id: int = db.Column(
        db.Integer,
        db.ForeignKey(
            "users.id", ondelete="CASCADE", onupdate="CASCADE"
        ),
        nullable=False
    )
    """
    The ID of the user associated with this notification setting
    """

    user: User = db.relationship(
        "User",
        backref=db.backref(
            "notification_settings", lazy=True, cascade="all,delete"
        )
    )
    """
    The user associated with this notification setting
    """

    notification_type: str = \
        db.Column(db.Enum(NotificationType), nullable=False)
    """
    The notification type
    """

    minimum_score: int = db.Column(db.Integer, default=0)
    """
    The minimum score for notification items
    """

    value: bool = db.Column(db.Boolean, nullable=False, default=False)
    """
    Whether or not the notification is active or not
    """

    def __json__(self, include_children: bool = False) -> Dict[str, Any]:
        """
        Generates a dictionary containing the information of this model
        :param include_children: Specifies if children data models
                                 will be included or if they're limited to IDs
        :return: A dictionary representing the model's values
        """
        data = {
            "id": self.id,
            "notification_type": self.notification_type,
            "value": self.value
        }
        if include_children:
            data["user"] = self.user.__json__(include_children)
        return data



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import List
from flask.blueprints import Blueprint
from puffotter.flask.routes.decorators import api
from puffotter.flask.exceptions import ApiException
from otaku_info.Config import Config
from otaku_info.db.MediaId import MediaId
from otaku_info.db.MediaItem import MediaItem
from otaku_info.utils.enums import MediaType, ListService


def define_blueprint(blueprint_name: str) -> Blueprint:
    """
    Defines the blueprint for this route
    :param blueprint_name: The name of the blueprint
    :return: The blueprint
    """
    blueprint = Blueprint(blueprint_name, __name__)
    api_base_path = f"/api/v{Config.API_VERSION}"

    @blueprint.route(
        f"{api_base_path}/media_ids/<_service>/<_media_type>/<service_id>",
        methods=["GET"]
    )
    @api
    def media_ids(_service: str, _media_type: str, service_id: str):
        """
        Retrieves all media IDs for a media ID
        :return: The response
        """
        service = ListService(_service)
        media_type = MediaType(_media_type)

        matching_ids: List[MediaItem] = [
            x for x in
            MediaId.query
            .filter_by(service_id=service_id, service=service).all()
            if x.media_item.media_type == media_type
        ]

        if len(matching_ids) < 1:
            raise ApiException("ID does not exist", 404)

        media_item = matching_ids[0].media_item

        id_mappings = {
            x.service.value: x.service_id
            for x in MediaId.query.filter_by(media_item_id=media_item.id).all()
        }
        id_mappings["otaku_info"] = str(media_item.id)

        return id_mappings

    return blueprint



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from flask.blueprints import Blueprint
from typing import List, Tuple, Callable
from otaku_info.routes.manga import define_blueprint as __manga
from otaku_info.routes.external_service import define_blueprint \
    as __external_service
from otaku_info.routes.api.media_api import define_blueprint as __media_api
from otaku_info.routes.notifications import define_blueprint as \
    __notifications
from otaku_info.routes.media import define_blueprint as __media
from otaku_info.routes.ln import define_blueprint as __ln

blueprint_generators: List[Tuple[Callable[[str], Blueprint], str]] = [
    (__external_service, "external_service"),
    (__manga, "manga"),
    (__media_api, "media_api"),
    (__notifications, "notifications"),
    (__media, "media"),
    (__ln, "ln")
]
"""
Defines the functions used to create the various blueprints
as well as their names
"""



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Dict, Type
from otaku_info.utils.enums import ListService


mangadex_external_id_names: Dict[ListService, str] = {
    ListService.ANILIST: "al",
    ListService.ANIMEPLANET: "ap",
    ListService.KITSU: "kt",
    ListService.MANGAUPDATES: "mu",
    ListService.MYANIMELIST: "mal",
    ListService.MANGADEX: "NONE"
}
"""
The names used by mangadex to identify IDs from other services
"""


list_service_url_formats: Dict[ListService, str] = {
    ListService.ANILIST: "https://anilist.co/@{media_type}/@{id}",
    ListService.ANIMEPLANET: "https://www.anime-planet.com/"
                             "@{media_type}/@{id}",
    ListService.KITSU: "https://kitsu.io/anime/@{id}",
    ListService.MANGAUPDATES: "https://www.mangaupdates.com/"
                              "series.html?id=@{id}",
    ListService.MYANIMELIST: "https://myanimelist.net/@{media_type}/@{id}",
    ListService.MANGADEX: "https://mangadex.org/title/@{id}"
}
"""
Schemas for URLs for external services
"""


list_service_id_types: Dict[ListService, Type] = {
    ListService.ANILIST: int,
    ListService.ANIMEPLANET: str,
    ListService.KITSU: int,
    ListService.MANGAUPDATES: int,
    ListService.MYANIMELIST: int,
    ListService.MANGADEX: int
}
"""
Which type a list service ID should have
"""



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import time
from typing import List, Dict, Optional, Tuple
from puffotter.flask.base import db, app
from puffotter.flask.db.User import User
from otaku_info.db.MediaId import MediaId
from otaku_info.db.MediaItem import MediaItem
from otaku_info.db.MediaList import MediaList
from otaku_info.db.MediaListItem import MediaListItem
from otaku_info.db.MediaUserState import MediaUserState
from otaku_info.db.ServiceUsername import ServiceUsername
from otaku_info.utils.anilist.AnilistItem import AnilistUserItem, AnilistItem
from otaku_info.utils.anilist.api import load_anilist
from otaku_info.utils.enums import ListService, MediaType, MediaSubType


def fetch_anilist_data():
    """
    Retrieves all entries on the anilists of all users that provided
    an anilist username
    :return: None
    """
    start = time.time()
    app.logger.debug("Starting Anilist Update")
    usernames: List[ServiceUsername] = \
        ServiceUsername.query.filter_by(service=ListService.ANILIST).all()
    anilist_data = {
        user: {
            media_type: load_anilist(user.username, media_type)
            for media_type in MediaType
        }
        for user in usernames
    }
    media_items, media_ids, media_user_states, media_lists, media_list_items\
        = load_existing()

    app.logger.debug("Updating Media Entries")
    update_media_entries(
        anilist_data, media_items, media_ids
    )
    app.logger.debug("Updating Media User States")
    update_media_user_entries(
        anilist_data, media_items, media_ids, media_user_states
    )
    app.logger.debug("Updating Media Lists")
    update_media_lists(
        anilist_data,
        media_items,
        media_ids,
        media_user_states,
        media_lists,
        media_list_items
    )
    app.logger.info(f"Completed anilist update in {time.time() - start}")


def update_media_entries(
        anilist_data: Dict[
            ServiceUsername,
            Dict[MediaType, List[AnilistUserItem]]
        ],
        media_items: Dict[Tuple[str, MediaType, MediaSubType, str], MediaItem],
        media_ids: Dict[Tuple[ListService, int], MediaId]
):
    """
    Updates the media entries and anilist IDs
    :param anilist_data: The anilist data to store
    :param media_items: The preloaded media items
    :param media_ids: The preloaded media IDs
    :return: None
    """
    updated_ids: List[Tuple[ListService, int]] = []
    updated_items: List[Tuple[str, MediaType, MediaSubType, str]] = []

    for media_type in MediaType:

        anilist_entries: List[AnilistUserItem] = []
        for data in anilist_data.values():
            anilist_entries += data[media_type]

        for anilist_entry in anilist_entries:
            item_tuple, media_item = fetch_media_item(
                anilist_entry, media_items
            )
            if item_tuple not in updated_items:
                media_item = update_media_item(anilist_entry, media_item)
                media_items[item_tuple] = media_item
                updated_items.append(item_tuple)

            id_tuple, media_id = fetch_media_id(
                anilist_entry, media_items, media_ids, media_item
            )
            assert id_tuple is not None

            if id_tuple not in updated_ids:
                media_item = media_items[item_tuple]
                media_id = update_media_id(anilist_entry, media_item, media_id)
                media_ids[id_tuple] = media_id
                updated_ids.append(id_tuple)

    db.session.commit()
    return media_ids


def update_media_user_entries(
        anilist_data: Dict[
            ServiceUsername,
            Dict[MediaType, List[AnilistUserItem]]
        ],
        media_items: Dict[Tuple[str, MediaType, MediaSubType, str], MediaItem],
        media_ids: Dict[Tuple[ListService, int], MediaId],
        media_user_states: Dict[Tuple[int, int], MediaUserState]
):
    """
    Updates the individual users' current state for media items in
    thei ranilist account.
    :param anilist_data: The anilist data to enter into the database
    :param media_items: Preloaded media items
    :param media_ids: Preloaded media IDs
    :param media_user_states: Preloaded media user states
    :return: None
    """
    updated: List[Tuple[int, int]] = []

    for service_user, anilist in anilist_data.items():
        user_states = {
            x: y for x, y in media_user_states.items()
            if y.user_id == service_user.user_id
        }

        for media_type, anilist_entries in anilist.items():
            for entry in anilist_entries:
                id_tuple, media_id = \
                    fetch_media_id(entry, media_items, media_ids)
                assert media_id is not None

                user_state_id = (media_id.id, service_user.user_id)

                if user_state_id in updated:
                    continue

                media_user_state = media_user_states.get(user_state_id)

                media_user_state = update_media_user_state(
                    entry, media_id, service_user.user, media_user_state
                )

                updated.append(user_state_id)
                media_user_states[user_state_id] = media_user_state

        for user_state_tuple, user_state in user_states.items():
            if user_state_tuple not in updated:
                db.session.delete(user_state)
                media_user_states.pop(user_state_tuple)

        db.session.commit()


def update_media_lists(
        anilist_data: Dict[
            ServiceUsername,
            Dict[MediaType, List[AnilistUserItem]]
        ],
        media_items: Dict[Tuple[str, MediaType, MediaSubType, str], MediaItem],
        media_ids: Dict[Tuple[ListService, int], MediaId],
        media_user_states: Dict[Tuple[int, int], MediaUserState],
        media_lists: Dict[Tuple[str, int, ListService, MediaType], MediaList],
        media_list_items: Dict[Tuple[int, int], MediaListItem]
):
    """
    Updates the database for anilist user lists.
    This includes custom anilist lists.
    :param anilist_data: The anilist data to enter into the database
    :param media_items: Preloaded media items
    :param media_ids: Preloaded media IDs
    :param media_user_states: The current media user states in the database
    :param media_lists: The media lists currently in the database
    :param media_list_items: The media list items currently in the database
    :return: None
    """
    list_tuples_to_remove = list(media_lists.keys())
    list_item_tuples_to_remove = list(media_list_items.keys())

    for service_user, anilist in anilist_data.items():
        for media_type, entries in anilist.items():
            for entry in entries:

                list_tuple = (
                    entry.list_name,
                    service_user.user_id,
                    ListService.ANILIST,
                    media_type
                )
                if list_tuple in list_tuples_to_remove:
                    list_tuples_to_remove.remove(list_tuple)
                media_list = media_lists.get(list_tuple)

                if media_list is None:
                    media_list = MediaList(
                        user_id=service_user.user_id,
                        name=entry.list_name,
                        service=ListService.ANILIST,
                        media_type=media_type
                    )
                    db.session.add(media_list)
                    db.session.commit()
                    media_lists[list_tuple] = media_list

                _, media_id = fetch_media_id(entry, media_items, media_ids)
                assert media_id is not None

                state_tuple = (media_id.id, service_user.user_id)
                media_user_state = media_user_states[state_tuple]

                list_item_tuple = (media_list.id, media_user_state.id)
                if list_item_tuple in list_item_tuples_to_remove:
                    list_item_tuples_to_remove.remove(list_item_tuple)

                if list_item_tuple not in media_list_items:
                    list_item = MediaListItem(
                        media_list=media_list,
                        media_user_state=media_user_state
                    )
                    db.session.add(list_item)

            db.session.commit()

    for list_tuple in list_tuples_to_remove:
        if list_tuple in media_lists:
            db.session.delete(media_lists.pop(list_tuple))
    for list_item_tuple in list_item_tuples_to_remove:
        if list_item_tuple in media_list_items:
            db.session.delete(media_list_items.pop(list_item_tuple))

    db.session.commit()


def load_existing() -> Tuple[
    Dict[Tuple[str, MediaType, MediaSubType, str], MediaItem],
    Dict[Tuple[ListService, int], MediaId],
    Dict[Tuple[int, int], MediaUserState],
    Dict[Tuple[str, int, ListService, MediaType], MediaList],
    Dict[Tuple[int, int], MediaListItem]
]:
    """
    Loads current database contents, mapped to unique identifer tuples
    :return: The database contents
    """
    app.logger.debug("Loading Existing data for anilist update")
    media_items: Dict[Tuple[str, MediaType, MediaSubType, str], MediaItem] = {
        (x.romaji_title, x.media_type, x.media_subtype, x.cover_url): x
        for x in MediaItem.query.all()
    }
    app.logger.debug("Finished loading MediaItems")
    media_ids: Dict[Tuple[ListService, int], MediaId] = {
        (x.service, x.media_item_id): x
        for x in MediaId.query.all()
    }
    app.logger.debug("Finished loading MediaIds")
    media_user_states: Dict[Tuple[int, int], MediaUserState] = {
        (x.media_id_id, x.user_id): x
        for x in MediaUserState.query.all()
    }
    app.logger.debug("Finished loading MediaUserStates")
    media_lists: Dict[Tuple[str, int, ListService, MediaType], MediaList] = {
        (x.name, x.user_id, x.service, x.media_type): x
        for x in MediaList.query.all()
    }
    app.logger.debug("Finished loading MediaLists")
    media_list_items: Dict[Tuple[int, int], MediaListItem] = {
        (x.media_list_id, x.media_user_state_id): x
        for x in MediaListItem.query.all()
    }
    app.logger.debug("Finished loading MediaListItems")
    return media_items, media_ids, media_user_states, \
        media_lists, media_list_items


def fetch_media_item(
        anilist_entry: AnilistUserItem,
        media_items: Dict[Tuple[str, MediaType, MediaSubType, str], MediaItem],
) -> Tuple[
    Tuple[str, MediaType, MediaSubType, str],
    Optional[MediaItem]
]:
    """
    Retrieves an existing media item based on anilist data
    :param anilist_entry: The anilist entry to use
    :param media_items: The preloaded media items
    :return: The media item, or None if none exists
    """
    item_tuple = (
        anilist_entry.romaji_title,
        anilist_entry.media_type,
        anilist_entry.media_subtype,
        anilist_entry.cover_url
    )
    return item_tuple, media_items.get(item_tuple)


def fetch_media_id(
        anilist_entry: AnilistUserItem,
        media_items: Dict[Tuple[str, MediaType, MediaSubType, str], MediaItem],
        media_ids: Dict[Tuple[ListService, int], MediaId],
        media_item: Optional[MediaId] = None
) -> Tuple[Optional[Tuple[ListService, int]], Optional[MediaItem]]:
    """
    Retrieves an existing media ID based on anilist data
    :param anilist_entry: The anilist entry to use
    :param media_items: The preloaded media items
    :param media_ids: The preloaded media IDs
    :param media_item: Optional media item associated with the ID.
                       If not provided, will figure out using anilist data
    :return: The media ID, or None if none exists
    """
    if media_item is None:
        _, media_item = fetch_media_item(anilist_entry, media_items)
    if media_item is None:
        return None, None
    else:
        id_tuple = (
            ListService.ANILIST,
            media_item.id
        )
        return id_tuple, media_ids.get(id_tuple)


def update_media_item(
        new_data: AnilistItem,
        existing: Optional[MediaItem]
) -> MediaItem:
    """
    Updates or creates MediaItem database entries based on anilist data
    :param new_data: The new anilist data
    :param existing: The existing database entry. If None, will be created
    :return: The updated/created MediaItem object
    """
    media_item = MediaItem() if existing is None else existing
    media_item.media_type = new_data.media_type
    media_item.media_subtype = new_data.media_subtype
    media_item.english_title = new_data.english_title
    media_item.romaji_title = new_data.romaji_title
    media_item.cover_url = new_data.cover_url
    media_item.latest_release = new_data.latest_release
    media_item.releasing_state = new_data.releasing_state

    if existing is None:
        db.session.add(media_item)
        db.session.commit()
    return media_item


def update_media_id(
        new_data: AnilistItem,
        media_item: MediaItem,
        existing: Optional[MediaId]
) -> MediaId:
    """
    Updates/Creates a MediaId database entry based on anilist data
    :param new_data: The anilist data to use
    :param media_item: The media item associated with the ID
    :param existing: The existing database entry. If None, will be created
    :return: The updated/created MediaId object
    """
    media_id = MediaId() if existing is None else existing
    media_id.media_item = media_item
    media_id.service = ListService.ANILIST
    media_id.service_id = str(new_data.anilist_id)

    if existing is None:
        db.session.add(media_id)
        db.session.commit()
    return media_id


def update_media_user_state(
        new_data: AnilistUserItem,
        media_id: MediaId,
        user: User,
        existing: Optional[MediaUserState]
) -> MediaUserState:
    """
    Updates or creates a MediaUserState entry in the database
    :param new_data: The new anilist data
    :param media_id: The media ID of the anilist media item
    :param user: The user associated with the data
    :param existing: The existing database entry. If None, will be created
    :return: The updated/created MediaUserState object
    """
    media_user_state = MediaUserState() if existing is None else existing
    media_user_state.media_id = media_id
    media_user_state.consuming_state = new_data.consuming_state
    media_user_state.score = new_data.score
    media_user_state.progress = new_data.progress
    media_user_state.user = user

    if existing is None:
        db.session.add(media_user_state)
        db.session.commit()
    return media_user_state



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Optional
from otaku_info.utils.enums import ListService, MediaType
from otaku_info.utils.mappings import list_service_url_formats


def build_service_url(
        media_type: MediaType,
        service: ListService,
        service_id: str
) -> str:
    """
    Builds an URL for an external service based on an ID
    :param media_type: The media type for which to generate an URL
    :param service: The service for which to create the URL
    :param service_id: The ID of the media item on that service
    :return: The generated URL
    """
    url_format = list_service_url_formats[service]
    url = url_format \
        .replace("@{media_type}", media_type.value) \
        .replace("@{id}", service_id)
    return url


def build_title(english_title: Optional[str], romaji_title: str) -> str:
    """
    Determines the title based on an English and Romaji title
    :param english_title: The english title
    :param romaji_title: The romaji title
    :return: The title
    """
    return romaji_title if english_title is None else english_title



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from flask import request, render_template, redirect, url_for
from flask.blueprints import Blueprint
from flask_login import current_user, login_required
from puffotter.flask.base import db
from puffotter.flask.db.TelegramChatId import TelegramChatId
from otaku_info.utils.enums import NotificationType
from otaku_info.db.NotificationSetting import NotificationSetting


def define_blueprint(blueprint_name: str) -> Blueprint:
    """
    Defines a Blueprint that handles notifications for users
    :param blueprint_name: The name of the blueprint
    :return: The blueprint
    """
    blueprint = Blueprint(blueprint_name, __name__)

    @blueprint.route("/notifications", methods=["GET"])
    @login_required
    def notifications():
        """
        Displays the notification settings page
        :return: The response
        """
        telegram_chat_id = \
            TelegramChatId.query.filter_by(user=current_user).first()
        settings = {
            x: False
            for x in NotificationType
        }
        settings.update({
            x.notification_type: x
            for x in
            NotificationSetting.query.filter_by(user_id=current_user.id).all()
        })
        return render_template(
            "user_management/notifications.html",
            telegram_chat_id=telegram_chat_id,
            notification_settings=settings
        )

    @blueprint.route("/set_notification_settings", methods=["POST"])
    @login_required
    def set_notification_settings():
        """
        Sets the notification settings
        :return: Redirect to notifications page
        """
        existing_settings = {
            x.notification_type: x
            for x in
            NotificationSetting.query.filter_by(user_id=current_user.id).all()
        }
        for notification_type in NotificationType:
            active_input = request.form.get(notification_type.value, "off")
            active_value = active_input == "on"

            min_score_text = request.form.get(
                notification_type.value + "_min_score",
                "0"
            )
            try:
                min_score = int(min_score_text)
            except IndexError:
                min_score = 0

            setting = existing_settings.get(notification_type)
            if setting is None:
                setting = NotificationSetting(
                    user=current_user,
                    notification_type=notification_type,
                    minimum_score=min_score
                )
                db.session.add(setting)

            setting.value = active_value
            setting.minimum_score = min_score
            db.session.commit()

        return redirect(url_for("notifications.notifications"))

    return blueprint



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Dict, Any, Optional
from puffotter.flask.base import db
from puffotter.flask.db.ModelMixin import ModelMixin
from otaku_info.utils.db_model_helper import build_title
from otaku_info.utils.enums import ReleasingState, MediaType, MediaSubType


class MediaItem(ModelMixin, db.Model):
    """
    Database model for media items.
    These model a generic, site-agnostic representation of a series.
    """

    __tablename__ = "media_items"
    """
    The name of the database table
    """

    __table_args__ = (
        db.UniqueConstraint(
            "media_type",
            "media_subtype",
            "romaji_title",
            "cover_url",
            name="unique_media_item"
        ),
    )
    """
    Makes sure that objects that should be unique are unique
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes the Model
        :param args: The constructor arguments
        :param kwargs: The constructor keyword arguments
        """
        super().__init__(*args, **kwargs)

    media_type: MediaType = db.Column(db.Enum(MediaType), nullable=False)
    """
    The media type of the list item
    """

    media_subtype: MediaSubType = db.Column(
        db.Enum(MediaSubType), nullable=False
    )
    """
    The subtype (for example, TV short, movie oneshot etc)
    """

    english_title: Optional[str] = db.Column(db.Unicode(255), nullable=True)
    """
    The English title of the media item
    """

    romaji_title: str = db.Column(db.Unicode(255), nullable=False)
    """
    The Japanese title of the media item written in Romaji
    """

    cover_url: str = db.Column(db.String(255), nullable=False)
    """
    An URL to a cover image of the media item
    """

    latest_release: Optional[int] = db.Column(db.Integer, nullable=True)
    """
    The latest release chapter/episode for this media item
    """

    releasing_state: ReleasingState = db.Column(
        db.Enum(ReleasingState), nullable=False
    )
    """
    The current releasing state of the media item
    """

    @property
    def title(self) -> str:
        """
        :return: The default title for the media item.
        """
        return build_title(self.english_title, self.romaji_title)

    def __json__(self, include_children: bool = False) -> Dict[str, Any]:
        """
        Generates a dictionary containing the information of this model
        :param include_children: Specifies if children data models
                                 will be included or if they're limited to IDs
        :return: A dictionary representing the model's values
        """
        data = {
            "id": self.id,
            "media_type": self.media_type.value,
            "media_subtype": self.media_subtype.value,
            "english_title": self.english_title,
            "romaji_title": self.romaji_title,
            "cover_url": self.cover_url,
            "latest_release": self.latest_release,
            "releasing_state": self.releasing_state.value
        }
        return data



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from unittest import TestCase
from otaku_info.utils.enums import ListService
from otaku_info.utils.mappings import list_service_id_types, \
    list_service_url_formats, mangadex_external_id_names


class TestMappings(TestCase):
    """
    Class that tests enum mappings
    """

    def test_completeness(self):
        """
        Tests that mappings include all possible enum types
        :return: None
        """
        for list_service in ListService:
            self.assertTrue(list_service in list_service_id_types)
            self.assertTrue(list_service in list_service_url_formats)
            self.assertTrue(list_service in mangadex_external_id_names)



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from datetime import datetime
from flask import request, render_template, redirect, url_for
from flask.blueprints import Blueprint
from otaku_info.db.LnRelease import LnRelease
from otaku_info.utils.dates import MONTHS, map_month_name_to_month_number, \
    map_month_number_to_month_name


def define_blueprint(blueprint_name: str) -> Blueprint:
    """
    Defines the blueprint for this route
    :param blueprint_name: The name of the blueprint
    :return: The blueprint
    """
    blueprint = Blueprint(blueprint_name, __name__)

    @blueprint.route("/ln/releases", methods=["GET"])
    def ln_releases():
        """
        Displays light novel releases
        :return: The response
        """
        try:
            year = int(request.args.get("year"))
        except (TypeError, ValueError):
            year = None
        try:
            month = int(request.args["month"])
        except KeyError:
            month = None
        except (TypeError, ValueError):
            month_string = request.args["month"]
            if month_string.lower() == "all":
                month = None
            else:
                month = map_month_name_to_month_number(month_string)

        now = datetime.utcnow()
        if not (year is not None and month is None):
            if year is None:
                year = now.year
            if month is None:
                month = now.month

        all_releases = LnRelease.query.all()
        years = list(set([x.release_date.year for x in all_releases]))
        years.sort()

        releases = [
            x for x in all_releases
            if x.release_date.year == year
        ]
        if month is not None:
            releases = [x for x in releases if x.release_date.month == month]
        releases.sort(key=lambda x: x.release_date)

        if month is None:
            month_name = "all"
        else:
            month_name = map_month_number_to_month_name(month)

        return render_template(
            "ln/ln_releases.html",
            releases=releases,
            years=years,
            months=MONTHS + ["all"],
            selected_year=year,
            selected_month=month_name
        )

    @blueprint.route("/ln/releases", methods=["POST"])
    def ln_releases_form():
        """
        Handles form requests and forwards it to the appropriate GET URL.
        """
        year = request.form.get("year")
        month = request.form.get("month")
        get_url = url_for("ln.ln_releases") + f"?year={year}&month={month}"
        return redirect(get_url)

    return blueprint



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import json
import requests
from typing import Dict, Optional
from otaku_info.utils.enums import ListService
from otaku_info.utils.mappings import mangadex_external_id_names, \
    list_service_id_types


def get_external_ids(mangadex_id: int) -> Optional[Dict[ListService, str]]:
    """
    Retrieves associated IDs for a mangadex ID
    :param mangadex_id: The mangadex ID
    :return: The other IDs, mapped to their list service
    """
    endpoint = "https://mangadex.org/api/manga/{}".format(mangadex_id)
    response = json.loads(requests.get(endpoint).text)

    ids = {ListService.MANGADEX: str(mangadex_id)}

    if response["status"] != "OK":
        return None
    else:
        links = response["manga"]["links"]
        if links is None:
            return ids

        for service, identifier in mangadex_external_id_names.items():
            if identifier in links:
                _id = links[identifier]
                id_type = list_service_id_types[service]

                if id_type == int:
                    _id = "".join([x for x in _id if x.isdigit()])

                try:
                    ids[service] = str(id_type(_id))
                except ValueError:
                    pass

    return ids
